# encoding: utf-8

"""Unit test suite for `cr.cube.matrix.measure` module."""

import numpy as np
import pytest

from cr.cube.cube import Cube
from cr.cube.dimension import Dimension
from cr.cube.matrix.cubemeasure import (
    CubeMeasures,
    _BaseUnweightedCubeCounts,
    _BaseWeightedCubeCounts,
)
from cr.cube.matrix.measure import (
    _BaseSecondOrderMeasure,
    _ColumnProportions,
    _ColumnShareSum,
    _ColumnUnweightedBases,
    _ColumnWeightedBases,
    _RowProportions,
    _RowShareSum,
    _RowUnweightedBases,
    _RowWeightedBases,
    SecondOrderMeasures,
    _Sums,
    _TableUnweightedBases,
    _TableWeightedBases,
    _TotalShareSum,
    _UnweightedCounts,
    _WeightedCounts,
)

from ...unitutil import ANY, class_mock, instance_mock, property_mock


class DescribeSecondOrderMeasures(object):
    """Unit test suite for `cr.cube.matrix.measure.SecondOrderMeasures` object."""

    @pytest.mark.parametrize(
        "measure_prop_name, MeasureCls",
        (
            ("column_proportions", _ColumnProportions),
            ("column_unweighted_bases", _ColumnUnweightedBases),
            ("column_weighted_bases", _ColumnWeightedBases),
            ("row_proportions", _RowProportions),
            ("row_unweighted_bases", _RowUnweightedBases),
            ("row_weighted_bases", _RowWeightedBases),
            ("table_unweighted_bases", _TableUnweightedBases),
            ("table_weighted_bases", _TableWeightedBases),
            ("weighted_counts", _WeightedCounts),
            ("unweighted_counts", _UnweightedCounts),
        ),
    )
    def it_provides_access_to_various_measure_objects(
        self,
        request,
        dimensions_,
        _cube_measures_prop_,
        cube_measures_,
        measure_prop_name,
        MeasureCls,
    ):
        measure_ = instance_mock(request, MeasureCls)
        MeasureCls_ = class_mock(
            request,
            "cr.cube.matrix.measure.%s" % MeasureCls.__name__,
            return_value=measure_,
        )
        _cube_measures_prop_.return_value = cube_measures_
        measures = SecondOrderMeasures(None, dimensions_, None)

        measure = getattr(measures, measure_prop_name)

        MeasureCls_.assert_called_once_with(dimensions_, measures, cube_measures_)
        assert measure is measure_

    def it_provides_access_to_the_columns_pruning_base(
        self, _cube_measures_prop_, cube_measures_, unweighted_cube_counts_
    ):
        _cube_measures_prop_.return_value = cube_measures_
        cube_measures_.unweighted_cube_counts = unweighted_cube_counts_
        unweighted_cube_counts_.columns_pruning_base = np.array([8, 5, 7, 4])
        measures = SecondOrderMeasures(None, None, None)

        assert measures.columns_pruning_base.tolist() == [8, 5, 7, 4]

    def it_provides_access_to_the_rows_pruning_base(
        self, _cube_measures_prop_, cube_measures_, unweighted_cube_counts_
    ):
        _cube_measures_prop_.return_value = cube_measures_
        cube_measures_.unweighted_cube_counts = unweighted_cube_counts_
        unweighted_cube_counts_.rows_pruning_base = np.array([7, 4, 0, 2])
        measures = SecondOrderMeasures(None, None, None)

        assert measures.rows_pruning_base.tolist() == [7, 4, 0, 2]

    def it_provides_access_to_the_cube_measures_to_help(
        self, request, cube_, dimensions_, cube_measures_
    ):
        CubeMeasures_ = class_mock(
            request,
            "cr.cube.matrix.measure.CubeMeasures",
            return_value=cube_measures_,
        )
        measures = SecondOrderMeasures(cube_, dimensions_, slice_idx=42)

        cube_measures = measures._cube_measures

        CubeMeasures_.assert_called_once_with(cube_, dimensions_, 42)
        assert cube_measures is cube_measures_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def cube_measures_(self, request):
        return instance_mock(request, CubeMeasures)

    @pytest.fixture
    def _cube_measures_prop_(self, request):
        return property_mock(request, SecondOrderMeasures, "_cube_measures")

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))

    @pytest.fixture
    def unweighted_cube_counts_(self, request):
        return instance_mock(request, _BaseUnweightedCubeCounts)


class Describe_BaseSecondOrderMeasure(object):
    """Unit test suite for `cr.cube.matrix.measure._BaseSecondOrderMeasure` object."""

    def it_assembles_the_blocks_for_the_measure(self, request):
        property_mock(
            request, _BaseSecondOrderMeasure, "_base_values", return_value="A"
        )
        property_mock(
            request, _BaseSecondOrderMeasure, "_subtotal_columns", return_value="B"
        )
        property_mock(
            request, _BaseSecondOrderMeasure, "_subtotal_rows", return_value="C"
        )
        property_mock(
            request, _BaseSecondOrderMeasure, "_intersections", return_value="D"
        )
        measure = _BaseSecondOrderMeasure(None, None, None)

        blocks = measure.blocks

        assert blocks == [["A", "B"], ["C", "D"]]

    def it_provides_access_to_the_unweighted_cube_counts_object_to_help(
        self, request, cube_measures_
    ):
        unweighted_cube_counts_ = instance_mock(request, _BaseUnweightedCubeCounts)
        cube_measures_.unweighted_cube_counts = unweighted_cube_counts_
        measure = _BaseSecondOrderMeasure(None, None, cube_measures_)

        assert measure._unweighted_cube_counts is unweighted_cube_counts_

    def it_provides_access_to_the_weighted_cube_counts_object_to_help(
        self, request, cube_measures_
    ):
        weighted_cube_counts_ = instance_mock(request, _BaseWeightedCubeCounts)
        cube_measures_.weighted_cube_counts = weighted_cube_counts_
        measure = _BaseSecondOrderMeasure(None, None, cube_measures_)

        weighted_cube_counts = measure._weighted_cube_counts

        assert weighted_cube_counts is weighted_cube_counts_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_measures_(self, request):
        return instance_mock(request, CubeMeasures)


class Describe_ColumnProportions(object):
    """Unit test suite for `cr.cube.matrix.measure._ColumnProportions` object."""

    def it_computes_its_blocks(self, request):
        SumSubtotals_ = class_mock(request, "cr.cube.matrix.measure.SumSubtotals")
        SumSubtotals_.blocks.return_value = [[5.0, 12.0], [21.0, 32.0]]
        column_weighted_bases_ = instance_mock(
            request, _ColumnWeightedBases, blocks=[[5.0, 6.0], [7.0, 8.0]]
        )
        second_order_measures_ = instance_mock(
            request,
            SecondOrderMeasures,
            column_weighted_bases=column_weighted_bases_,
        )
        cube_measures_ = class_mock(request, "cr.cube.matrix.cubemeasure.CubeMeasures")

        column_proportions = _ColumnProportions(
            None, second_order_measures_, cube_measures_
        )

        assert column_proportions.blocks == [[1.0, 2.0], [3.0, 4.0]]
        SumSubtotals_.blocks.assert_called_once_with(ANY, None, diff_cols_nan=True)


class Describe_ColumnUnweightedBases(object):
    """Unit test suite for `cr.cube.matrix.measure._ColumnUnweightedBases` object."""

    def it_computes_its_base_values_to_help(
        self, _unweighted_cube_counts_prop_, unweighted_cube_counts_
    ):
        _unweighted_cube_counts_prop_.return_value = unweighted_cube_counts_
        unweighted_cube_counts_.column_bases = np.arange(6).reshape(2, 3)
        column_unweighted_bases = _ColumnUnweightedBases(None, None, None)

        assert column_unweighted_bases._base_values.tolist() == [[0, 1, 2], [3, 4, 5]]

    def it_computes_its_intersections_block_to_help(
        self, request, _base_values_prop_, dimensions_, SumSubtotals_
    ):
        property_mock(
            request,
            _ColumnUnweightedBases,
            "_subtotal_columns",
            return_value=np.array([[9, 6], [9, 6], [9, 6]]),
        )
        _base_values_prop_.return_value = [[9, 8, 7], [6, 5, 4], [3, 2, 1]]
        SumSubtotals_.intersections.return_value = np.array([[8, 4], [3, 7]])
        column_unweighted_bases = _ColumnUnweightedBases(dimensions_, None, None)

        intersections = column_unweighted_bases._intersections

        SumSubtotals_.intersections.assert_called_once_with(
            [[9, 8, 7], [6, 5, 4], [3, 2, 1]], dimensions_, diff_cols_nan=True
        )
        assert intersections.tolist() == [[9, 6], [9, 6]]

    def it_computes_its_subtotal_columns_to_help(
        self, _base_values_prop_, dimensions_, SumSubtotals_
    ):
        _base_values_prop_.return_value = [[1, 2], [3, 4]]
        SumSubtotals_.subtotal_columns.return_value = np.array([[5, 8], [3, 7]])
        column_unweighted_bases = _ColumnUnweightedBases(dimensions_, None, None)

        subtotal_columns = column_unweighted_bases._subtotal_columns

        SumSubtotals_.subtotal_columns.assert_called_once_with(
            [[1, 2], [3, 4]], dimensions_, diff_cols_nan=True
        )
        assert subtotal_columns.tolist() == [[5, 8], [3, 7]]

    def it_computes_its_subtotal_rows_to_help(
        self,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        _unweighted_cube_counts_prop_,
        unweighted_cube_counts_,
    ):
        _base_values_prop_.return_value = [[4, 3], [2, 1]]
        SumSubtotals_.subtotal_rows.return_value = np.array([[8, 3], [6, 4]])
        _unweighted_cube_counts_prop_.return_value = unweighted_cube_counts_
        unweighted_cube_counts_.columns_base = np.array([4, 7])
        column_unweighted_bases = _ColumnUnweightedBases(dimensions_, None, None)

        subtotal_rows = column_unweighted_bases._subtotal_rows

        SumSubtotals_.subtotal_rows.assert_called_once_with(
            [[4, 3], [2, 1]], dimensions_, diff_cols_nan=True
        )
        assert subtotal_rows.tolist() == [[4, 7], [4, 7]]

    def but_it_returns_empty_array_of_right_shape_when_there_are_no_row_subtotals(
        self, _base_values_prop_, dimensions_, SumSubtotals_
    ):
        """Empty shape must be (0, ncols) to compose properly in `np.block()` call."""
        _base_values_prop_.return_value = [[4, 3, 2], [1, 0, 9]]
        SumSubtotals_.subtotal_rows.return_value = np.array([], dtype=int).reshape(0, 3)
        column_unweighted_bases = _ColumnUnweightedBases(dimensions_, None, None)

        subtotal_rows = column_unweighted_bases._subtotal_rows

        SumSubtotals_.subtotal_rows.assert_called_once_with(
            [[4, 3, 2], [1, 0, 9]], dimensions_, diff_cols_nan=True
        )
        assert subtotal_rows.tolist() == []
        assert subtotal_rows.shape == (0, 3)
        assert subtotal_rows.dtype == int

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _base_values_prop_(self, request):
        return property_mock(request, _ColumnUnweightedBases, "_base_values")

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))

    @pytest.fixture
    def SumSubtotals_(self, request):
        return class_mock(request, "cr.cube.matrix.measure.SumSubtotals")

    @pytest.fixture
    def unweighted_cube_counts_(self, request):
        return instance_mock(request, _BaseUnweightedCubeCounts)

    @pytest.fixture
    def _unweighted_cube_counts_prop_(self, request):
        return property_mock(request, _ColumnUnweightedBases, "_unweighted_cube_counts")


class Describe_ColumnWeightedBases(object):
    """Unit test suite for `cr.cube.matrix.measure._ColumnWeightedBases` object."""

    def it_computes_its_base_values_to_help(
        self, _weighted_cube_counts_prop_, weighted_cube_counts_
    ):
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        weighted_cube_counts_.column_bases = np.arange(6).reshape(3, 2)
        column_weighted_bases = _ColumnWeightedBases(None, None, None)

        assert column_weighted_bases._base_values.tolist() == [[0, 1], [2, 3], [4, 5]]

    def it_computes_its_intersections_block_to_help(
        self, request, _base_values_prop_, dimensions_, SumSubtotals_
    ):
        property_mock(
            request,
            _ColumnWeightedBases,
            "_subtotal_columns",
            return_value=np.array([[9.9, 6.6], [9.9, 6.6], [9.9, 6.6]]),
        )
        _base_values_prop_.return_value = [
            [9.9, 8.8, 7.7],
            [6.6, 5.5, 4.4],
            [3.3, 2.2, 1.1],
        ]
        SumSubtotals_.intersections.return_value = np.array([[8.8, 4.4], [3.3, 7.7]])
        column_weighted_bases = _ColumnWeightedBases(dimensions_, None, None)

        intersections = column_weighted_bases._intersections

        SumSubtotals_.intersections.assert_called_once_with(
            [[9.9, 8.8, 7.7], [6.6, 5.5, 4.4], [3.3, 2.2, 1.1]],
            dimensions_,
            diff_cols_nan=True,
        )
        assert intersections.tolist() == [[9.9, 6.6], [9.9, 6.6]]

    def it_computes_its_subtotal_columns_to_help(
        self, _base_values_prop_, dimensions_, SumSubtotals_
    ):
        _base_values_prop_.return_value = [[1.1, 2.2], [3.3, 4.4]]
        SumSubtotals_.subtotal_columns.return_value = np.array([[5.5, 8.8], [3.3, 7.7]])
        column_weighted_bases = _ColumnWeightedBases(dimensions_, None, None)

        subtotal_columns = column_weighted_bases._subtotal_columns

        SumSubtotals_.subtotal_columns.assert_called_once_with(
            [[1.1, 2.2], [3.3, 4.4]], dimensions_, diff_cols_nan=True
        )
        assert subtotal_columns.tolist() == [[5.5, 8.8], [3.3, 7.7]]

    def it_computes_its_subtotal_rows_to_help(
        self,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        _weighted_cube_counts_prop_,
        weighted_cube_counts_,
    ):
        _base_values_prop_.return_value = [[4.4, 3.3], [2.2, 1.1]]
        SumSubtotals_.subtotal_rows.return_value = np.array([[8.8, 3.3], [6.6, 4.4]])
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        weighted_cube_counts_.columns_margin = np.array([4.4, 7.7])
        column_weighted_bases = _ColumnWeightedBases(dimensions_, None, None)

        subtotal_rows = column_weighted_bases._subtotal_rows

        SumSubtotals_.subtotal_rows.assert_called_once_with(
            [[4.4, 3.3], [2.2, 1.1]], dimensions_, diff_cols_nan=True
        )
        assert subtotal_rows.tolist() == [[4.4, 7.7], [4.4, 7.7]]

    def but_it_returns_empty_array_of_right_shape_when_there_are_no_row_subtotals(
        self, _base_values_prop_, dimensions_, SumSubtotals_
    ):
        """Empty shape must be (0, ncols) to compose properly in `np.block()` call."""
        _base_values_prop_.return_value = [[4.4, 3.3, 2.2], [1.1, 0.0, 9.9]]
        SumSubtotals_.subtotal_rows.return_value = np.array([], dtype=int).reshape(0, 3)
        column_weighted_bases = _ColumnWeightedBases(dimensions_, None, None)

        subtotal_rows = column_weighted_bases._subtotal_rows

        SumSubtotals_.subtotal_rows.assert_called_once_with(
            [[4.4, 3.3, 2.2], [1.1, 0.0, 9.9]], dimensions_, diff_cols_nan=True
        )
        assert subtotal_rows.tolist() == []
        assert subtotal_rows.shape == (0, 3)
        assert subtotal_rows.dtype == int

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _base_values_prop_(self, request):
        return property_mock(request, _ColumnWeightedBases, "_base_values")

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))

    @pytest.fixture
    def SumSubtotals_(self, request):
        return class_mock(request, "cr.cube.matrix.measure.SumSubtotals")

    @pytest.fixture
    def weighted_cube_counts_(self, request):
        return instance_mock(request, _BaseWeightedCubeCounts)

    @pytest.fixture
    def _weighted_cube_counts_prop_(self, request):
        return property_mock(request, _ColumnWeightedBases, "_weighted_cube_counts")


class Describe_RowProportions(object):
    """Unit test suite for `cr.cube.matrix.measure._RowProportions` object."""

    def it_computes_its_blocks(self, request):
        SumSubtotals_ = class_mock(request, "cr.cube.matrix.measure.SumSubtotals")
        SumSubtotals_.blocks.return_value = [[5.0, 12.0], [21.0, 32.0]]
        row_weighted_bases_ = instance_mock(
            request, _RowWeightedBases, blocks=[[5.0, 6.0], [7.0, 8.0]]
        )
        second_order_measures_ = instance_mock(
            request,
            SecondOrderMeasures,
            row_weighted_bases=row_weighted_bases_,
        )
        cube_measures_ = class_mock(request, "cr.cube.matrix.cubemeasure.CubeMeasures")

        row_proportions = _RowProportions(None, second_order_measures_, cube_measures_)

        assert row_proportions.blocks == [[1.0, 2.0], [3.0, 4.0]]
        SumSubtotals_.blocks.assert_called_once_with(ANY, None, diff_rows_nan=True)


class Describe_RowUnweightedBases(object):
    """Unit test suite for `cr.cube.matrix.measure._RowUnweightedBases` object."""

    def it_computes_its_base_values_to_help(
        self, _unweighted_cube_counts_prop_, unweighted_cube_counts_
    ):
        _unweighted_cube_counts_prop_.return_value = unweighted_cube_counts_
        unweighted_cube_counts_.row_bases = np.arange(6).reshape(2, 3)
        row_unweighted_bases = _RowUnweightedBases(None, None, None)

        assert row_unweighted_bases._base_values.tolist() == [[0, 1, 2], [3, 4, 5]]

    @pytest.mark.parametrize(
        "subtotal_rows, intersections, expected_value, expected_shape",
        (
            (
                np.array([[9, 9, 9], [6, 6, 6]]),
                np.array([[8, 4], [3, 7]]),
                [[9, 9], [6, 6]],
                (2, 2),
            ),
            (
                np.array([], dtype=int).reshape(0, 3),
                np.array([], dtype=int).reshape(0, 2),
                [],
                (0, 2),
            ),
        ),
    )
    def it_computes_its_intersections_block_to_help(
        self,
        request,
        subtotal_rows,
        intersections,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        expected_value,
        expected_shape,
    ):
        property_mock(
            request, _RowUnweightedBases, "_subtotal_rows", return_value=subtotal_rows
        )
        _base_values_prop_.return_value = [[9, 8, 7], [6, 5, 4], [3, 2, 1]]
        SumSubtotals_.intersections.return_value = intersections
        row_unweighted_bases = _RowUnweightedBases(dimensions_, None, None)

        intersections = row_unweighted_bases._intersections

        SumSubtotals_.intersections.assert_called_once_with(
            [[9, 8, 7], [6, 5, 4], [3, 2, 1]], dimensions_, diff_rows_nan=True
        )
        assert intersections.tolist() == expected_value
        assert intersections.shape == expected_shape
        assert intersections.dtype == int

    def it_computes_its_subtotal_columns_to_help(
        self,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        _unweighted_cube_counts_prop_,
        unweighted_cube_counts_,
    ):
        _base_values_prop_.return_value = [[3, 4, 5], [1, 2, 3]]
        SumSubtotals_.subtotal_columns.return_value = np.array([[3, 8], [4, 6]])
        _unweighted_cube_counts_prop_.return_value = unweighted_cube_counts_
        unweighted_cube_counts_.rows_base = np.array([7, 4])
        row_unweighted_bases = _RowUnweightedBases(dimensions_, None, None)

        subtotal_columns = row_unweighted_bases._subtotal_columns

        SumSubtotals_.subtotal_columns.assert_called_once_with(
            [[3, 4, 5], [1, 2, 3]], dimensions_, diff_rows_nan=True
        )
        assert subtotal_columns.tolist() == [[7, 7], [4, 4]]

    def but_it_returns_empty_array_of_right_shape_when_there_are_no_column_subtotals(
        self, _base_values_prop_, dimensions_, SumSubtotals_
    ):
        """Empty shape must be (nrows, 0) to compose properly in `np.block()` call."""
        _base_values_prop_.return_value = [[2, 3, 4], [9, 0, 1]]
        SumSubtotals_.subtotal_columns.return_value = np.array([], dtype=int).reshape(
            3, 0
        )
        row_unweighted_bases = _RowUnweightedBases(dimensions_, None, None)

        subtotal_columns = row_unweighted_bases._subtotal_columns

        SumSubtotals_.subtotal_columns.assert_called_once_with(
            [[2, 3, 4], [9, 0, 1]], dimensions_, diff_rows_nan=True
        )
        assert subtotal_columns.tolist() == [[], [], []]
        assert subtotal_columns.shape == (3, 0)
        assert subtotal_columns.dtype == int

    def it_computes_its_subtotal_rows_to_help(
        self, _base_values_prop_, dimensions_, SumSubtotals_
    ):
        _base_values_prop_.return_value = [[1, 2], [3, 4]]
        SumSubtotals_.subtotal_rows.return_value = np.array([[5, 8], [3, 7]])
        row_unweighted_bases = _RowUnweightedBases(dimensions_, None, None)

        subtotal_rows = row_unweighted_bases._subtotal_rows

        SumSubtotals_.subtotal_rows.assert_called_once_with(
            [[1, 2], [3, 4]], dimensions_, diff_rows_nan=True
        )
        assert subtotal_rows.tolist() == [[5, 8], [3, 7]]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _base_values_prop_(self, request):
        return property_mock(request, _RowUnweightedBases, "_base_values")

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))

    @pytest.fixture
    def SumSubtotals_(self, request):
        return class_mock(request, "cr.cube.matrix.measure.SumSubtotals")

    @pytest.fixture
    def unweighted_cube_counts_(self, request):
        return instance_mock(request, _BaseUnweightedCubeCounts)

    @pytest.fixture
    def _unweighted_cube_counts_prop_(self, request):
        return property_mock(request, _RowUnweightedBases, "_unweighted_cube_counts")


class Describe_RowWeightedBases(object):
    """Unit test suite for `cr.cube.matrix.measure._RowWeightedBases` object."""

    def it_computes_its_base_values_to_help(
        self, _weighted_cube_counts_prop_, weighted_cube_counts_
    ):
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        weighted_cube_counts_.row_bases = np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]])
        row_weighted_bases = _RowWeightedBases(None, None, None)

        assert row_weighted_bases._base_values.tolist() == [
            [1.1, 2.2, 3.3],
            [4.4, 5.5, 6.6],
        ]

    @pytest.mark.parametrize(
        "subtotal_rows, intersections, expected_value, expected_shape",
        (
            (
                np.array([[9.9, 9.9, 9.9], [6.6, 6.6, 6.6]]),
                np.array([[8.8, 4.4], [3.3, 7.7]]),
                [[9.9, 9.9], [6.6, 6.6]],
                (2, 2),
            ),
            (
                np.array([], dtype=float).reshape(0, 3),
                np.array([], dtype=float).reshape(0, 2),
                [],
                (0, 2),
            ),
        ),
    )
    def it_computes_its_intersections_block_to_help(
        self,
        request,
        subtotal_rows,
        intersections,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        expected_value,
        expected_shape,
    ):
        property_mock(
            request, _RowWeightedBases, "_subtotal_rows", return_value=subtotal_rows
        )
        _base_values_prop_.return_value = [
            [9.9, 8.8, 7.7],
            [6.6, 5.5, 4.4],
            [3.3, 2.2, 1.1],
        ]
        SumSubtotals_.intersections.return_value = intersections
        row_weighted_bases = _RowWeightedBases(dimensions_, None, None)

        intersections = row_weighted_bases._intersections

        SumSubtotals_.intersections.assert_called_once_with(
            [[9.9, 8.8, 7.7], [6.6, 5.5, 4.4], [3.3, 2.2, 1.1]], dimensions_
        )
        assert intersections.tolist() == expected_value
        assert intersections.shape == expected_shape
        assert intersections.dtype == float

    def it_computes_its_subtotal_columns_to_help(
        self,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        _weighted_cube_counts_prop_,
        weighted_cube_counts_,
    ):
        _base_values_prop_.return_value = [[3.3, 4.4, 5.5], [1.1, 2.2, 3.3]]
        SumSubtotals_.subtotal_columns.return_value = np.array([[3.3, 8.8], [4.4, 6.6]])
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        weighted_cube_counts_.rows_margin = np.array([7.7, 4.4])
        row_weighted_bases = _RowWeightedBases(dimensions_, None, None)

        subtotal_columns = row_weighted_bases._subtotal_columns

        SumSubtotals_.subtotal_columns.assert_called_once_with(
            [[3.3, 4.4, 5.5], [1.1, 2.2, 3.3]], dimensions_, diff_rows_nan=True
        )
        assert subtotal_columns.tolist() == [[7.7, 7.7], [4.4, 4.4]]

    def but_it_returns_empty_array_of_right_shape_when_there_are_no_column_subtotals(
        self, _base_values_prop_, dimensions_, SumSubtotals_
    ):
        """Empty shape must be (nrows, 0) to compose properly in `np.block()` call."""
        _base_values_prop_.return_value = [[2.2, 3.3, 4.4], [9.9, 0.0, 1.1]]
        SumSubtotals_.subtotal_columns.return_value = np.array([], dtype=int).reshape(
            3, 0
        )
        row_weighted_bases = _RowWeightedBases(dimensions_, None, None)

        subtotal_columns = row_weighted_bases._subtotal_columns

        SumSubtotals_.subtotal_columns.assert_called_once_with(
            [[2.2, 3.3, 4.4], [9.9, 0.0, 1.1]], dimensions_, diff_rows_nan=True
        )
        assert subtotal_columns.tolist() == [[], [], []]
        assert subtotal_columns.shape == (3, 0)
        assert subtotal_columns.dtype == int

    def it_computes_its_subtotal_rows_to_help(
        self, _base_values_prop_, dimensions_, SumSubtotals_
    ):
        _base_values_prop_.return_value = [[1.1, 2.2], [3.3, 4.4]]
        SumSubtotals_.subtotal_rows.return_value = np.array([[5.5, 8.8], [3.3, 7.7]])
        row_weighted_bases = _RowWeightedBases(dimensions_, None, None)

        subtotal_rows = row_weighted_bases._subtotal_rows

        SumSubtotals_.subtotal_rows.assert_called_once_with(
            [[1.1, 2.2], [3.3, 4.4]], dimensions_, diff_rows_nan=True
        )
        assert subtotal_rows.tolist() == [[5.5, 8.8], [3.3, 7.7]]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _base_values_prop_(self, request):
        return property_mock(request, _RowWeightedBases, "_base_values")

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))

    @pytest.fixture
    def SumSubtotals_(self, request):
        return class_mock(request, "cr.cube.matrix.measure.SumSubtotals")

    @pytest.fixture
    def weighted_cube_counts_(self, request):
        return instance_mock(request, _BaseWeightedCubeCounts)

    @pytest.fixture
    def _weighted_cube_counts_prop_(self, request):
        return property_mock(request, _RowWeightedBases, "_weighted_cube_counts")


class Describe_ColumnShareSum(object):
    """Unit test suite for `cr.cube.matrix.measure._ColumnShareSum` object."""

    def it_computes_its_blocks(self, request):
        SumSubtotals_ = class_mock(request, "cr.cube.matrix.measure.SumSubtotals")
        SumSubtotals_.blocks.return_value = [
            [[5.0, 12.0], [21.0, 32.0]],
            [[], []],
            [[], []],
            [[], []],
        ]
        sums_blocks_ = instance_mock(request, _Sums, blocks=[[5.0, 6.0], [7.0, 8.0]])
        second_order_measures_ = instance_mock(
            request,
            SecondOrderMeasures,
            sums=sums_blocks_,
        )
        cube_measures_ = class_mock(request, "cr.cube.matrix.cubemeasure.CubeMeasures")

        col_share_sum = _ColumnShareSum(None, second_order_measures_, cube_measures_)

        assert col_share_sum.blocks[0][0] == pytest.approx([0.2941176, 0.7058823])
        assert col_share_sum.blocks[0][1] == pytest.approx([0.3962264, 0.6037735])
        SumSubtotals_.blocks.assert_called_once_with(ANY, None, True, True)


class Describe_RowShareSum(object):
    """Unit test suite for `cr.cube.matrix.measure._RowShareSum` object."""

    def it_computes_its_blocks(self, request):
        SumSubtotals_ = class_mock(request, "cr.cube.matrix.measure.SumSubtotals")
        SumSubtotals_.blocks.return_value = [
            np.array([[[5.0, 12.0]], [[21.0, 32.0]]]),
            np.array([[[]], [[]]]),
            np.array([[[]], [[]]]),
            np.array([[[]], [[]]]),
        ]
        sums_blocks_ = instance_mock(
            request, _Sums, blocks=np.array([[[5.0, 6.0]], [[7.0, 8.0]]])
        )
        second_order_measures_ = instance_mock(
            request,
            SecondOrderMeasures,
            sums=sums_blocks_,
        )
        cube_measures_ = class_mock(request, "cr.cube.matrix.cubemeasure.CubeMeasures")

        row_share_sum = _RowShareSum(None, second_order_measures_, cube_measures_)

        assert row_share_sum.blocks[0][0] == pytest.approx(
            np.array([[0.29411765, 0.70588235]])
        )
        assert row_share_sum.blocks[0][1] == pytest.approx(
            np.array([[0.3962264, 0.6037735]])
        )
        SumSubtotals_.blocks.assert_called_once_with(ANY, None, True, True)


class Describe_TotalShareSum(object):
    """Unit test suite for `cr.cube.matrix.measure._RowShareSum` object."""

    def it_computes_its_blocks(self, request):
        SumSubtotals_ = class_mock(request, "cr.cube.matrix.measure.SumSubtotals")
        SumSubtotals_.blocks.return_value = [
            np.array([[[5.0, 12.0]], [[21.0, 32.0]]]),
            np.array([[[]], [[]]]),
            np.array([[[]], [[]]]),
            np.array([[[]], [[]]]),
        ]
        sums_blocks_ = instance_mock(
            request, _Sums, blocks=np.array([[[5.0, 6.0]], [[7.0, 8.0]]])
        )
        second_order_measures_ = instance_mock(
            request,
            SecondOrderMeasures,
            sums=sums_blocks_,
        )
        cube_measures_ = class_mock(request, "cr.cube.matrix.cubemeasure.CubeMeasures")

        total_share_sum = _TotalShareSum(None, second_order_measures_, cube_measures_)

        assert total_share_sum.blocks[0][0] == pytest.approx(
            np.array([[0.29411765, 0.70588235]])
        )
        assert total_share_sum.blocks[0][1] == pytest.approx(
            np.array([[0.3962264, 0.6037735]])
        )
        SumSubtotals_.blocks.assert_called_once_with(ANY, None, True, True)


class Describe_TableUnweightedBases(object):
    """Unit test suite for `cr.cube.matrix.measure._TableUnweightedBases` object."""

    def it_computes_its_base_values_to_help(
        self, _unweighted_cube_counts_prop_, unweighted_cube_counts_
    ):
        _unweighted_cube_counts_prop_.return_value = unweighted_cube_counts_
        unweighted_cube_counts_.table_bases = np.array([[3, 0, 5], [1, 4, 2]])
        table_unweighted_bases = _TableUnweightedBases(None, None, None)

        assert table_unweighted_bases._base_values.tolist() == [[3, 0, 5], [1, 4, 2]]

    @pytest.mark.parametrize(
        "intersections, table_base, expected_value, expected_shape",
        (
            # --- CAT_X_CAT with both row-subtotals and column-subtotals ---
            (
                np.array([[8, 4], [3, 7]]),
                7,
                [[7, 7], [7, 7]],
                (2, 2),
            ),
            # --- No intersections but table-base is an array (one of the MR cases) ---
            (
                np.array([], dtype=int).reshape(0, 2),
                [6, 6, 6],
                [],
                (0, 2),
            ),
            # --- No intersections and table-base is a scalar ---
            (
                np.array([], dtype=int).reshape(0, 2),
                9,
                [],
                (0, 2),
            ),
        ),
    )
    def it_computes_its_intersections_block_to_help(
        self,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        _unweighted_cube_counts_prop_,
        unweighted_cube_counts_,
        intersections,
        table_base,
        expected_value,
        expected_shape,
    ):
        _base_values_prop_.return_value = [[6, 5, 4], [9, 8, 7], [3, 2, 1]]
        SumSubtotals_.intersections.return_value = intersections
        _unweighted_cube_counts_prop_.return_value = unweighted_cube_counts_
        unweighted_cube_counts_.table_base = table_base
        table_unweighted_bases = _TableUnweightedBases(dimensions_, None, None)

        intersections = table_unweighted_bases._intersections

        SumSubtotals_.intersections.assert_called_once_with(
            [[6, 5, 4], [9, 8, 7], [3, 2, 1]], dimensions_
        )
        assert intersections.tolist() == expected_value
        assert intersections.shape == expected_shape
        assert intersections.dtype == int

    @pytest.mark.parametrize(
        "table_base, expected_value",
        ((np.array([7, 4]), [[7, 7], [4, 4]]), (9, [[9, 9], [9, 9]])),
    )
    def it_computes_its_subtotal_columns_to_help(
        self,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        _unweighted_cube_counts_prop_,
        unweighted_cube_counts_,
        table_base,
        expected_value,
    ):
        _base_values_prop_.return_value = [[5, 6, 3], [4, 1, 2]]
        SumSubtotals_.subtotal_columns.return_value = np.array([[0, 0], [0, 0]])
        _unweighted_cube_counts_prop_.return_value = unweighted_cube_counts_
        unweighted_cube_counts_.table_base = table_base
        table_unweighted_bases = _TableUnweightedBases(dimensions_, None, None)

        subtotal_columns = table_unweighted_bases._subtotal_columns

        SumSubtotals_.subtotal_columns.assert_called_once_with(
            [[5, 6, 3], [4, 1, 2]], dimensions_
        )
        assert subtotal_columns.tolist() == expected_value

    def but_it_returns_empty_array_of_right_shape_when_there_are_no_column_subtotals(
        self, _base_values_prop_, dimensions_, SumSubtotals_
    ):
        """Empty shape must be (nrows, 0) to compose properly in `np.block()` call."""
        _base_values_prop_.return_value = [[2, 3], [4, 9], [0, 1]]
        SumSubtotals_.subtotal_columns.return_value = np.array([], dtype=int).reshape(
            3, 0
        )
        table_unweighted_bases = _TableUnweightedBases(dimensions_, None, None)

        subtotal_columns = table_unweighted_bases._subtotal_columns

        SumSubtotals_.subtotal_columns.assert_called_once_with(
            [[2, 3], [4, 9], [0, 1]], dimensions_
        )
        assert subtotal_columns.tolist() == [[], [], []]
        assert subtotal_columns.shape == (3, 0)
        assert subtotal_columns.dtype == int

    @pytest.mark.parametrize(
        "subtotal_rows, table_base, expected_value, expected_shape",
        (
            # --- no subtotal-rows case (including MR_X_CAT and MR_X_MR cases) ---
            (np.array([], dtype=int).reshape(0, 2), None, [], (0, 2)),
            # --- scalar table-base (CAT_X_CAT case) ---
            (np.array([[0, 0], [0, 0]]), 8, [[8, 8], [8, 8]], (2, 2)),
            # --- vector table-base (CAT_X_MR case) ---
            (np.array([[0, 0], [0, 0]]), np.array([4, 2]), [[4, 2], [4, 2]], (2, 2)),
        ),
    )
    def it_computes_its_subtotal_rows_to_help(
        self,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        _unweighted_cube_counts_prop_,
        unweighted_cube_counts_,
        subtotal_rows,
        table_base,
        expected_value,
        expected_shape,
    ):
        _base_values_prop_.return_value = [[3, 5, 6], [2, 4, 1]]
        SumSubtotals_.subtotal_rows.return_value = subtotal_rows
        _unweighted_cube_counts_prop_.return_value = unweighted_cube_counts_
        unweighted_cube_counts_.table_base = table_base
        table_unweighted_bases = _TableUnweightedBases(dimensions_, None, None)

        subtotal_rows = table_unweighted_bases._subtotal_rows

        SumSubtotals_.subtotal_rows.assert_called_once_with(
            [[3, 5, 6], [2, 4, 1]], dimensions_
        )
        assert subtotal_rows.tolist() == expected_value
        assert subtotal_rows.dtype == int
        assert subtotal_rows.shape == expected_shape

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _base_values_prop_(self, request):
        return property_mock(request, _TableUnweightedBases, "_base_values")

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))

    @pytest.fixture
    def SumSubtotals_(self, request):
        return class_mock(request, "cr.cube.matrix.measure.SumSubtotals")

    @pytest.fixture
    def unweighted_cube_counts_(self, request):
        return instance_mock(request, _BaseUnweightedCubeCounts)

    @pytest.fixture
    def _unweighted_cube_counts_prop_(self, request):
        return property_mock(request, _TableUnweightedBases, "_unweighted_cube_counts")


class Describe_TableWeightedBases(object):
    """Unit test suite for `cr.cube.matrix.measure._TableWeightedBases` object."""

    def it_computes_its_base_values_to_help(
        self, _weighted_cube_counts_prop_, weighted_cube_counts_
    ):
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        weighted_cube_counts_.table_bases = np.array([[3.3, 0.0, 5.5], [1.1, 4.4, 2.2]])
        table_weighted_bases = _TableWeightedBases(None, None, None)

        assert table_weighted_bases._base_values.tolist() == [
            [3.3, 0.0, 5.5],
            [1.1, 4.4, 2.2],
        ]

    @pytest.mark.parametrize(
        "intersections, table_margin, expected_value, expected_shape",
        (
            # --- CAT_X_CAT with both row-subtotals and column-subtotals ---
            (
                np.array([[8, 4], [3, 7]]),
                7,
                [[7, 7], [7, 7]],
                (2, 2),
            ),
            # --- All other cases, intersections array is empty ---
            (
                np.array([], dtype=int).reshape(0, 2),
                None,
                [],
                (0, 2),
            ),
        ),
    )
    def it_computes_its_intersections_block_to_help(
        self,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        _weighted_cube_counts_prop_,
        weighted_cube_counts_,
        intersections,
        table_margin,
        expected_value,
        expected_shape,
    ):
        _base_values_prop_.return_value = [
            [6.6, 5.5, 4.4],
            [9.9, 8.8, 7.7],
            [3.3, 2.2, 1.1],
        ]
        SumSubtotals_.intersections.return_value = intersections
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        weighted_cube_counts_.table_margin = table_margin
        table_weighted_bases = _TableWeightedBases(dimensions_, None, None)

        intersections = table_weighted_bases._intersections

        SumSubtotals_.intersections.assert_called_once_with(
            [[6.6, 5.5, 4.4], [9.9, 8.8, 7.7], [3.3, 2.2, 1.1]], dimensions_
        )
        assert intersections.tolist() == expected_value
        assert intersections.shape == expected_shape
        assert intersections.dtype == int

    @pytest.mark.parametrize(
        "subtotal_columns_, table_margin, expected_value",
        (
            # --- CAT_X_CAT case, scalar table-margin ---
            (
                np.array([[6.1, 6.2], [6.3, 6.4]]),
                9.9,
                np.array([[9.9, 9.9], [9.9, 9.9]]),
            ),
            # --- CAT_X_MR and MR_X_MR cases, subtotal-columns is (nrows, 0) array ---
            (
                np.array([[], []]),
                None,
                np.array([[], []]),
            ),
            # --- MR_X_CAT case, table-margin is a 1D array "column" ---
            (
                np.array([[6.6, 7.7], [8.8, 9.9]]),
                np.array([3.4, 5.6]),
                np.array([[3.4, 3.4], [5.6, 5.6]]),
            ),
        ),
    )
    def it_computes_its_subtotal_columns_to_help(
        self,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        subtotal_columns_,
        _weighted_cube_counts_prop_,
        weighted_cube_counts_,
        table_margin,
        expected_value,
    ):
        _base_values_prop_.return_value = [[5.5, 6.6, 3.3], [4.4, 1.1, 2.2]]
        SumSubtotals_.subtotal_columns.return_value = subtotal_columns_
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        weighted_cube_counts_.table_margin = table_margin
        table_weighted_bases = _TableWeightedBases(dimensions_, None, None)

        subtotal_columns = table_weighted_bases._subtotal_columns

        SumSubtotals_.subtotal_columns.assert_called_once_with(
            [[5.5, 6.6, 3.3], [4.4, 1.1, 2.2]], dimensions_
        )
        assert subtotal_columns == pytest.approx(expected_value)

    @pytest.mark.parametrize(
        "subtotal_rows_, table_margin, expected_value",
        (
            # --- CAT_X_CAT case, scalar table-margin ---
            (
                np.array([[6.1, 6.2, 6.3], [6.4, 6.5, 6.6]]),
                9.9,
                np.array([[9.9, 9.9, 9.9], [9.9, 9.9, 9.9]]),
            ),
            # --- CAT_X_MR case, table-margin is a 1D array "row" ---
            (
                np.array([[6.1, 6.2, 6.3], [6.4, 6.5, 6.6]]),
                np.array([3.4, 5.6, 7.8]),
                np.array(
                    [
                        [3.4, 5.6, 7.8],
                        [3.4, 5.6, 7.8],
                    ]
                ),
            ),
            # --- MR_X_CAT and MR_X_MR cases, subtotal-rows is (0, ncols) array ---
            (
                np.array([]).reshape(0, 3),
                None,
                np.array([]).reshape(0, 3),
            ),
        ),
    )
    def it_computes_its_subtotal_rows_to_help(
        self,
        _base_values_prop_,
        dimensions_,
        SumSubtotals_,
        subtotal_rows_,
        _weighted_cube_counts_prop_,
        weighted_cube_counts_,
        table_margin,
        expected_value,
    ):
        _base_values_prop_.return_value = [[5.5, 6.6, 3.3], [4.4, 1.1, 2.2]]
        SumSubtotals_.subtotal_rows.return_value = subtotal_rows_
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        weighted_cube_counts_.table_margin = table_margin
        table_weighted_bases = _TableWeightedBases(dimensions_, None, None)

        subtotal_rows = table_weighted_bases._subtotal_rows

        SumSubtotals_.subtotal_rows.assert_called_once_with(
            [[5.5, 6.6, 3.3], [4.4, 1.1, 2.2]], dimensions_
        )
        assert subtotal_rows == pytest.approx(expected_value)

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _base_values_prop_(self, request):
        return property_mock(request, _TableWeightedBases, "_base_values")

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))

    @pytest.fixture
    def SumSubtotals_(self, request):
        return class_mock(request, "cr.cube.matrix.measure.SumSubtotals")

    @pytest.fixture
    def weighted_cube_counts_(self, request):
        return instance_mock(request, _BaseWeightedCubeCounts)

    @pytest.fixture
    def _weighted_cube_counts_prop_(self, request):
        return property_mock(request, _TableWeightedBases, "_weighted_cube_counts")


class Describe_UnweightedCounts(object):
    """Unit test suite for `cr.cube.matrix.measure._UnweightedCounts` object."""

    def it_computes_its_blocks_to_help(self, request, dimensions_):
        # --- these need to be in list form because the assert-called-with mechanism
        # --- uses equality, which doesn't work on numpy arrays. Normally this would be
        # --- the array itself.
        ucounts = np.arange(12).reshape(3, 4).tolist()
        unweighted_cube_counts_ = instance_mock(
            request, _BaseUnweightedCubeCounts, unweighted_counts=ucounts
        )
        property_mock(
            request,
            _UnweightedCounts,
            "_unweighted_cube_counts",
            return_value=unweighted_cube_counts_,
        )
        SumSubtotals_ = class_mock(request, "cr.cube.matrix.measure.SumSubtotals")
        SumSubtotals_.blocks.return_value = [[[1], [2]], [[3], [4]]]
        unweighted_counts = _UnweightedCounts(dimensions_, None, None)

        blocks = unweighted_counts.blocks

        SumSubtotals_.blocks.assert_called_once_with(ucounts, dimensions_)
        assert blocks == [[[1], [2]], [[3], [4]]]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))


class Describe_WeightedCounts(object):
    """Unit test suite for `cr.cube.matrix.measure._WeightedCounts` object."""

    def it_computes_its_blocks_to_help(self, request, dimensions_):
        # --- these need to be in list form because the assert-called-with mechanism
        # --- uses equality, which doesn't work on numpy arrays. Normally this would be
        # --- the array itself.
        counts = np.arange(12).reshape(3, 4).tolist()
        weighted_cube_counts_ = instance_mock(
            request, _BaseWeightedCubeCounts, weighted_counts=counts
        )
        property_mock(
            request,
            _WeightedCounts,
            "_weighted_cube_counts",
            return_value=weighted_cube_counts_,
        )
        SumSubtotals_ = class_mock(request, "cr.cube.matrix.measure.SumSubtotals")
        SumSubtotals_.blocks.return_value = [[[1], [2]], [[3], [4]]]
        weighted_counts = _WeightedCounts(dimensions_, None, None)

        blocks = weighted_counts.blocks

        SumSubtotals_.blocks.assert_called_once_with(counts, dimensions_)
        assert blocks == [[[1], [2]], [[3], [4]]]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))



# encoding: utf-8

"""The `StripeAssembler` object provides the external interface for this module.

Its name derives from its role to "assemble" a finished 1D array ("stripe") for a
particular measure from the base measure values and inserted subtotals, to reorder the
rows according to the dimension *order* transforms, and to hide rows that are either
hidden by the user or "pruned" because they contain no observations.
"""

from __future__ import division

import numpy as np

from cr.cube.collator import ExplicitOrderCollator, PayloadOrderCollator
from cr.cube.enums import COLLATION_METHOD as CM
from cr.cube.stripe.measure import StripeMeasures
from cr.cube.util import lazyproperty


class StripeAssembler(object):
    """Provides measures, marginals, and totals for a (1D) strand cube-slice.

    An assembled stripe measure is a 1D ndarray reflecting all ordering, insertion, and
    hiding transforms applied to the dimension. An assembled margin is often a scalar.

    `cube` is the `cr.cube.Cube` object containing the data for this matrix.

    `rows_dimension` is the Dimension object describing the stripe.

    `slice_idx` is an int offset indicating which portion of `cube` data to use for this
    stripe.
    """

    def __init__(self, cube, rows_dimension, ca_as_0th, slice_idx):
        self._cube = cube
        self._rows_dimension = rows_dimension
        self._ca_as_0th = ca_as_0th
        self._slice_idx = slice_idx

    @lazyproperty
    def inserted_row_idxs(self):
        """tuple of int index of each inserted row in this stripe.

        Provided index values correspond rows after any insertion of subtotals,
        re-ordering, and hiding/pruning.
        """
        return tuple(i for i, row_idx in enumerate(self._row_order) if row_idx < 0)

    @lazyproperty
    def means(self):
        """1D np.float64 ndarray of mean for each row.

        Raises ValueError when the cube-result does not include a means cube-measure.
        """
        return self._assemble_vector(self._measures.means.blocks)

    @lazyproperty
    def row_count(self):
        """int count of rows in this stripe.

        This count includes inserted rows but not rows that have been hidden/pruned.
        """
        return len(self._row_order)

    @lazyproperty
    def row_labels(self):
        """1D str ndarray of row name for each matrix row.

        These are suitable for use as row headings; labels for subtotal rows appear in
        the sequence and labels are ordered to correspond with their respective data
        row.
        """
        return np.array(
            [e.label for e in self._rows_dimension.valid_elements]
            + [s.label for s in self._rows_dimension.subtotals]
        )[self._row_order]

    @lazyproperty
    def rows_dimension_fills(self):
        """tuple of optional RGB str like "#def032" fill color for each strand row.

        Each value reflects the resolved element-fill transform cascade. The length and
        ordering of the sequence correspond to the rows in the slice, including
        accounting for insertions, ordering, and hidden rows. A fill value is `None`
        when no explicit fill color is defined for that row, indicating the default fill
        color for that row should be used, probably coming from a caller-defined theme.
        """
        element_fills = tuple(e.fill for e in self._rows_dimension.valid_elements)
        return tuple(
            (element_fills[idx] if idx > -1 else None) for idx in self._row_order
        )

    @lazyproperty
    def scale_mean(self):
        """Optional float mean of row numeric-values (scale).

        This value is `None` when no row-elements have numeric-value. The numeric value
        (aka. "scaled-count") for a row is its count multiplied by the numeric-value of
        its element. For example, if 100 women responded "Very Likely" and the
        numeric-value of the "Very Likely" response (element) was 4, then the
        scaled-count for that row would be 400. The scale mean is the average of those
        scale values over the total count of responses. The count of a row lacking a
        numeric value does not contribute to either the numerator or denominator of that
        computation.
        """
        return self._measures.scaled_counts.scale_mean

    @lazyproperty
    def scale_median(self):
        """Optional float/int median of scaled weighted-counts.

        This value is `None` when no rows have a numeric-value assigned. The median is
        equal to a count of 1 multiplied by one of the row numeric-values, so the value
        is equal to one of the assigned numeric-values (and retains its int/float type).
        """
        return self._measures.scaled_counts.scale_median

    @lazyproperty
    def scale_stddev(self):
        """Optional np.float64 standard-deviation of scaled weighted-counts.

        This value is `None` when no rows have a numeric-value assigned. The value has
        the same units as the assigned numeric values and indicates the dispersion of
        the scaled-count distribution from its mean (scale-mean).
        """
        return self._measures.scaled_counts.scale_stddev

    @lazyproperty
    def scale_stderr(self):
        """Optional np.float64 standard-error of scaled weighted counts.

        This value is `None` when no rows have a numeric-value assigned.
        """
        return self._measures.scaled_counts.scale_stderr

    @lazyproperty
    def share_sum(self):
        """1D np.float64 ndarray of share of sum for each row.

        Raises ValueError when the cube-result does not include a sum cube-measure.
        """
        return self._assemble_vector(self._measures.share_sum.blocks)

    @lazyproperty
    def stddev(self):
        """1D np.float64 ndarray of stddev for each row.

        Raises ValueError when the cube-result does not include a stddev cube-measure.
        """
        return self._assemble_vector(self._measures.stddev.blocks)

    @lazyproperty
    def sums(self):
        """1D np.float64 ndarray of sum for each row.

        Raises ValueError when the cube-result does not include a sum cube-measure.
        """
        return self._assemble_vector(self._measures.sums.blocks)

    @lazyproperty
    def table_base_range(self):
        """[min, max] np.float64 ndarray range of (total) unweighted-N for this stripe.

        A non-MR stripe will have a single base, represented by min and max being the
        same value. Each row of an MR stripe has a distinct base, which is reduced to a
        range in that case.
        """
        return self._measures.unweighted_bases.table_base_range

    @lazyproperty
    def table_margin_range(self):
        """[min, max] np.float64 ndarray range of (total) unweighted-N for this stripe.

        A non-MR stripe will have a single margin, represented by min and max being the
        same value. Each row of an MR stripe has a distinct margin, which is reduced to
        a min/max range in that case.
        """
        return self._measures.weighted_bases.table_margin_range

    @lazyproperty
    def table_proportion_stddevs(self):
        """1D np.float64 ndarray of table-proportion std-deviation for each row."""
        return self._assemble_vector(self._measures.table_proportion_stddevs.blocks)

    @lazyproperty
    def table_proportion_stderrs(self):
        """1D np.float64 ndarray of table-proportion std-error for each row."""
        return self._assemble_vector(self._measures.table_proportion_stderrs.blocks)

    @lazyproperty
    def table_proportions(self):
        """1D np.float64 ndarray of fraction of weighted-N contributed by each row."""
        return self._assemble_vector(self._measures.table_proportions.blocks)

    @lazyproperty
    def unweighted_bases(self):
        """1D np.float64 ndarray of (unweighted) table-proportion denominator per row."""
        return self._assemble_vector(self._measures.unweighted_bases.blocks)

    @lazyproperty
    def unweighted_counts(self):
        """1D np.float64 ndarray of unweighted count for each row of stripe."""
        return self._assemble_vector(self._measures.unweighted_counts.blocks)

    @lazyproperty
    def weighted_bases(self):
        """1D np.float64 ndarray of weighted table-proportion denominator per row."""
        return self._assemble_vector(self._measures.weighted_bases.blocks)

    @lazyproperty
    def weighted_counts(self):
        """1D np.float64 ndarray of weighted count for each row of stripe."""
        return self._assemble_vector(self._measures.weighted_counts.blocks)

    def _assemble_vector(self, blocks):
        """Return 1D ndarray of base_vector with inserted subtotals, in order.

        `blocks` is a pair of two 1D arrays, first the base-values and then the subtotal
        values of the stripe vector. The returned array is sequenced in the computed
        row order including possibly removing hidden or pruned values.
        """
        return np.concatenate(blocks)[self._row_order]

    @lazyproperty
    def _measures(self):
        """StripeMeasures collection object for this stripe."""
        return StripeMeasures(
            self._cube, self._rows_dimension, self._ca_as_0th, self._slice_idx
        )

    @lazyproperty
    def _row_order(self):
        """1D np.int64 ndarray of signed int idx for each assembled row of stripe.

        Negative values represent inserted subtotal-row locations. Indices appear in the
        order rows are to appear in the final result.
        """
        # --- specify dtype explicitly to prevent error when display-order is empty. The
        # --- default dtype is float, which cannot be used to index an array.
        return np.array(
            _BaseOrderHelper.display_order(self._rows_dimension, self._measures),
            dtype=int,
        )


# === ORDER HELPERS ===


class _BaseOrderHelper(object):
    """Base class for ordering helpers."""

    def __init__(self, rows_dimension, measures):
        self._rows_dimension = rows_dimension
        self._measures = measures

    @classmethod
    def display_order(cls, rows_dimension, measures):
        """1D np.int64 ndarray of signed int idx for each row of stripe.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        HelperCls = (
            _SortByValueHelper
            if rows_dimension.collation_method == CM.OPPOSING_ELEMENT
            else _OrderHelper
        )
        return HelperCls(rows_dimension, measures)._display_order

    @lazyproperty
    def _display_order(self):
        """tuple of signed int idx for each row of stripe.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        raise NotImplementedError(
            "`%s` must implement `._display_order`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def _empty_row_idxs(self):
        """tuple of int index for each row with N = 0.

        These rows are subject to pruning, depending on a user setting in the dimension.
        """
        return tuple(i for i, N in enumerate(self._measures.pruning_base) if N == 0)


class _OrderHelper(_BaseOrderHelper):
    """Encapsulates the complexity of the various kinds of row ordering."""

    @lazyproperty
    def _display_order(self):
        """tuple of signed int idx for each row of stripe.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        CollatorCls = (
            ExplicitOrderCollator
            if self._rows_dimension.collation_method == CM.EXPLICIT_ORDER
            else PayloadOrderCollator
        )
        return CollatorCls.display_order(self._rows_dimension, self._empty_row_idxs)


class _SortByValueHelper(_BaseOrderHelper):
    """Orders rows by their values."""



# encoding: utf-8

"""Unit test suite for `cr.cube.stripe.measure` module."""

import numpy as np
import pytest

from cr.cube.cube import Cube
from cr.cube.dimension import Dimension
from cr.cube.stripe.cubemeasure import (
    _BaseCubeMeans,
    _BaseUnweightedCubeCounts,
    _BaseWeightedCubeCounts,
    _CatUnweightedCubeCounts,
    _CatWeightedCubeCounts,
    CubeMeasures,
)
from cr.cube.stripe.measure import (
    _BaseSecondOrderMeasure,
    _Means,
    _ScaledCounts,
    StripeMeasures,
    _TableProportionStddevs,
    _TableProportionStderrs,
    _TableProportionVariances,
    _TableProportions,
    _UnweightedBases,
    _UnweightedCounts,
    _WeightedBases,
    _WeightedCounts,
)

from ...unitutil import class_mock, instance_mock, property_mock


class DescribeStripeMeasures(object):
    """Unit test suite for `cr.cube.stripe.measure.StripeMeasures` object."""

    @pytest.mark.parametrize(
        "measure_prop_name, MeasureCls",
        (
            ("means", _Means),
            ("scaled_counts", _ScaledCounts),
            ("table_proportion_stddevs", _TableProportionStddevs),
            ("table_proportion_stderrs", _TableProportionStderrs),
            ("table_proportion_variances", _TableProportionVariances),
            ("table_proportions", _TableProportions),
            ("unweighted_bases", _UnweightedBases),
            ("unweighted_counts", _UnweightedCounts),
            ("weighted_bases", _WeightedBases),
            ("weighted_counts", _WeightedCounts),
        ),
    )
    def it_provides_access_to_various_measure_objects(
        self,
        request,
        rows_dimension_,
        _cube_measures_prop_,
        cube_measures_,
        measure_prop_name,
        MeasureCls,
    ):
        measure_ = instance_mock(request, MeasureCls)
        MeasureCls_ = class_mock(
            request,
            "cr.cube.stripe.measure.%s" % MeasureCls.__name__,
            return_value=measure_,
        )
        _cube_measures_prop_.return_value = cube_measures_
        measures = StripeMeasures(None, rows_dimension_, None, None)

        measure = getattr(measures, measure_prop_name)

        MeasureCls_.assert_called_once_with(rows_dimension_, measures, cube_measures_)
        assert measure is measure_

    def it_provides_access_to_the_pruning_base(
        self, request, _cube_measures_prop_, cube_measures_
    ):
        unweighted_cube_counts_ = instance_mock(
            request, _BaseUnweightedCubeCounts, pruning_base=np.array([0, 2, 7])
        )
        cube_measures_.unweighted_cube_counts = unweighted_cube_counts_
        _cube_measures_prop_.return_value = cube_measures_
        measures = StripeMeasures(None, None, None, None)

        assert measures.pruning_base.tolist() == [0, 2, 7]

    def it_provides_access_to_the_cube_measures_to_help(
        self, request, cube_, rows_dimension_, cube_measures_
    ):
        CubeMeasures_ = class_mock(
            request,
            "cr.cube.stripe.measure.CubeMeasures",
            return_value=cube_measures_,
        )
        measures = StripeMeasures(cube_, rows_dimension_, True, slice_idx=42)

        cube_measures = measures._cube_measures

        CubeMeasures_.assert_called_once_with(cube_, rows_dimension_, True, 42)
        assert cube_measures is cube_measures_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def cube_measures_(self, request):
        return instance_mock(request, CubeMeasures)

    @pytest.fixture
    def _cube_measures_prop_(self, request):
        return property_mock(request, StripeMeasures, "_cube_measures")

    @pytest.fixture
    def rows_dimension_(self, request):
        return instance_mock(request, Dimension)


class Describe_BaseSecondOrderMeasure(object):
    """Unit test suite for `cr.cube.stripe.measure._BaseSecondOrderMeasure` object."""

    def it_gathers_the_blocks_for_the_measure(self, request):
        property_mock(request, _BaseSecondOrderMeasure, "base_values", return_value="A")
        property_mock(
            request, _BaseSecondOrderMeasure, "subtotal_values", return_value="B"
        )
        measure = _BaseSecondOrderMeasure(None, None, None)

        assert measure.blocks == ("A", "B")

    def it_provides_access_to_the_unweighted_cube_counts_object_to_help(
        self, request, cube_measures_
    ):
        unweighted_cube_counts_ = instance_mock(request, _BaseUnweightedCubeCounts)
        cube_measures_.unweighted_cube_counts = unweighted_cube_counts_
        measure = _BaseSecondOrderMeasure(None, None, cube_measures_)

        assert measure._unweighted_cube_counts is unweighted_cube_counts_

    def it_provides_access_to_the_weighted_cube_counts_object_to_help(
        self, request, cube_measures_
    ):
        weighted_cube_counts_ = instance_mock(request, _BaseWeightedCubeCounts)
        cube_measures_.weighted_cube_counts = weighted_cube_counts_
        measure = _BaseSecondOrderMeasure(None, None, cube_measures_)

        assert measure._weighted_cube_counts is weighted_cube_counts_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_measures_(self, request):
        return instance_mock(request, CubeMeasures)


class Describe_Means(object):
    """Unit test suite for `cr.cube.stripe.measure._Means` object."""

    def it_computes_its_base_values_to_help(self, request):
        cube_means_ = instance_mock(
            request, _BaseCubeMeans, means=np.array([1.1, 2.2, 3.3])
        )
        cube_measures_ = instance_mock(request, CubeMeasures, cube_means=cube_means_)
        means = _Means(None, None, cube_measures_)

        assert means.base_values == pytest.approx([1.1, 2.2, 3.3])

    def it_computes_its_subtotal_values_to_help(self, request):
        property_mock(request, _Means, "base_values", return_value=[1.1, 2.2, 3.3])
        rows_dimension_ = instance_mock(request, Dimension)
        NanSubtotals_ = class_mock(request, "cr.cube.stripe.measure.NanSubtotals")
        NanSubtotals_.subtotal_values.return_value = np.array([np.nan, np.nan])
        means = _Means(rows_dimension_, None, None)

        subtotal_values = means.subtotal_values

        NanSubtotals_.subtotal_values.assert_called_once_with(
            [1.1, 2.2, 3.3], rows_dimension_
        )
        assert subtotal_values == pytest.approx([np.nan, np.nan], nan_ok=True)


class Describe_ScaledCounts(object):
    """Unit test suite for `cr.cube.stripe.measure._ScaledCounts` object."""

    @pytest.mark.parametrize(
        "numeric_values, total_weighted_count, expected_value",
        (
            (np.array([]), 0, None),
            (np.array([1, 2, 3]), 0, None),
            (np.array([1, 2, 3]), 100, 4),
        ),
    )
    def it_knows_the_scale_mean(
        self,
        _numeric_values_prop_,
        numeric_values,
        _total_weighted_count_prop_,
        total_weighted_count,
        _total_scaled_count_prop_,
        expected_value,
    ):
        _numeric_values_prop_.return_value = numeric_values
        _total_weighted_count_prop_.return_value = total_weighted_count
        _total_scaled_count_prop_.return_value = 400
        scaled_counts = _ScaledCounts(None, None, None)

        assert scaled_counts.scale_mean == expected_value

    @pytest.mark.parametrize(
        "numeric_values, weighted_counts, expected_value",
        (
            (np.array([]), np.array([]), None),
            (np.array([3, 1, 2]), np.array([300, 100, 200]), 2.5),
        ),
    )
    def it_knows_the_scale_median(
        self,
        _numeric_values_prop_,
        numeric_values,
        _weighted_counts_prop_,
        weighted_counts,
        _total_weighted_count_prop_,
        expected_value,
    ):
        _numeric_values_prop_.return_value = numeric_values
        _weighted_counts_prop_.return_value = weighted_counts
        _total_weighted_count_prop_.return_value = np.sum(weighted_counts)
        scaled_counts = _ScaledCounts(None, None, None)

        assert scaled_counts.scale_median == expected_value

    @pytest.mark.parametrize(
        "numeric_values, scale_variance, expected_value",
        (
            (np.array([]), None, None),
            (np.array([1, 2, 3]), 4.0, 2.0),
        ),
    )
    def it_knows_the_scale_stddev(
        self,
        _numeric_values_prop_,
        numeric_values,
        _scale_variance_prop_,
        scale_variance,
        expected_value,
    ):
        _numeric_values_prop_.return_value = numeric_values
        _scale_variance_prop_.return_value = scale_variance
        scaled_counts = _ScaledCounts(None, None, None)

        assert scaled_counts.scale_stddev == expected_value

    @pytest.mark.parametrize(
        "numeric_values, scale_variance, total_weighted_count, expected_value",
        (
            (np.array([]), None, None, None),
            (np.array([1, 2, 3]), 4.0, 100.0, 0.2),
        ),
    )
    def it_knows_the_scale_stderr(
        self,
        _numeric_values_prop_,
        numeric_values,
        _scale_variance_prop_,
        scale_variance,
        _total_weighted_count_prop_,
        total_weighted_count,
        expected_value,
    ):
        _numeric_values_prop_.return_value = numeric_values
        _scale_variance_prop_.return_value = scale_variance
        _total_weighted_count_prop_.return_value = total_weighted_count
        scaled_counts = _ScaledCounts(None, None, None)

        assert scaled_counts.scale_stderr == expected_value

    def it_knows_which_elements_have_a_numeric_values_to_help(self, rows_dimension_):
        rows_dimension_.numeric_values = (1, np.nan, 3)
        scaled_counts = _ScaledCounts(rows_dimension_, None, None)

        assert scaled_counts._has_numeric_value.tolist() == [True, False, True]

    def it_gathers_the_numeric_values_to_help(
        self, rows_dimension_, _has_numeric_value_prop_
    ):
        rows_dimension_.numeric_values = (1, np.nan, 3)
        _has_numeric_value_prop_.return_value = np.array([True, False, True])
        scaled_counts = _ScaledCounts(rows_dimension_, None, None)

        assert scaled_counts._numeric_values.tolist() == [1, 3]

    def it_computes_the_scale_variance_to_help(
        self,
        request,
        _weighted_counts_prop_,
        _numeric_values_prop_,
        _total_weighted_count_prop_,
    ):
        _weighted_counts_prop_.return_value = np.array([100, 200, 300])
        _numeric_values_prop_.return_value = np.array([1, 2, 3])
        property_mock(request, _ScaledCounts, "scale_mean", return_value=2.333333)
        _total_weighted_count_prop_.return_value = 600
        scaled_counts = _ScaledCounts(None, None, None)

        assert scaled_counts._scale_variance == pytest.approx(0.5555556)

    def it_computes_the_total_scaled_count_to_help(
        self, _weighted_counts_prop_, _numeric_values_prop_
    ):
        _weighted_counts_prop_.return_value = np.array([10, 20, 30, 40])
        _numeric_values_prop_.return_value = np.array([1, 2, 3, 4])
        scaled_counts = _ScaledCounts(None, None, None)

        # --- 10 + 40 + 90 + 160 = 300 ---
        assert scaled_counts._total_scaled_count == 300

    def it_computes_the_total_weighted_count_to_help(self, _weighted_counts_prop_):
        _weighted_counts_prop_.return_value = np.array([10, 20, 30, 40])
        scaled_counts = _ScaledCounts(None, None, None)

        assert scaled_counts._total_weighted_count == 100

    def it_retrives_the_weighted_counts_to_help(
        self, request, _has_numeric_value_prop_
    ):
        weighted_cube_counts_ = instance_mock(
            request, _BaseWeightedCubeCounts, weighted_counts=np.array([1.1, 2.2, 3.3])
        )
        property_mock(
            request,
            _ScaledCounts,
            "_weighted_cube_counts",
            return_value=weighted_cube_counts_,
        )
        _has_numeric_value_prop_.return_value = np.array([True, False, True])
        scaled_counts = _ScaledCounts(None, None, None)

        assert scaled_counts._weighted_counts == pytest.approx([1.1, 3.3])

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _has_numeric_value_prop_(self, request):
        return property_mock(request, _ScaledCounts, "_has_numeric_value")

    @pytest.fixture
    def _numeric_values_prop_(self, request):
        return property_mock(request, _ScaledCounts, "_numeric_values")

    @pytest.fixture
    def rows_dimension_(self, request):
        return instance_mock(request, Dimension)

    @pytest.fixture
    def _scale_variance_prop_(self, request):
        return property_mock(request, _ScaledCounts, "_scale_variance")

    @pytest.fixture
    def _total_weighted_count_prop_(self, request):
        return property_mock(request, _ScaledCounts, "_total_weighted_count")

    @pytest.fixture
    def _total_scaled_count_prop_(self, request):
        return property_mock(request, _ScaledCounts, "_total_scaled_count")

    @pytest.fixture
    def _weighted_counts_prop_(self, request):
        return property_mock(request, _ScaledCounts, "_weighted_counts")


class Describe_TableProportionStddevs(object):
    """Unit test suite for `cr.cube.stripe.measure._TableProportionStddevs` object."""

    def it_computes_its_base_values_to_help(
        self, measures_, table_proportion_variances_
    ):
        table_proportion_variances_.base_values = np.array([0.04, 0.09, 0.16])
        measures_.table_proportion_variances = table_proportion_variances_
        table_proportion_stddevs = _TableProportionStddevs(None, measures_, None)

        assert table_proportion_stddevs.base_values == pytest.approx([0.2, 0.3, 0.4])

    def it_computes_its_subtotal_values_to_help(
        self, measures_, table_proportion_variances_
    ):
        table_proportion_variances_.subtotal_values = np.array([0.25, 0.36])
        measures_.table_proportion_variances = table_proportion_variances_
        table_proportion_stddevs = _TableProportionStddevs(None, measures_, None)

        assert table_proportion_stddevs.subtotal_values == pytest.approx([0.5, 0.6])

    # fixture components ---------------------------------------------

    @pytest.fixture
    def measures_(self, request):
        return instance_mock(request, StripeMeasures)

    @pytest.fixture
    def table_proportion_variances_(self, request):
        return instance_mock(request, _TableProportionVariances)


class Describe_TableProportionStderrs(object):
    """Unit test suite for `cr.cube.stripe.measure._TableProportionStderrs` object."""

    def it_computes_its_base_values_to_help(
        self, measures_, table_proportion_variances_, weighted_bases_
    ):
        table_proportion_variances_.base_values = np.array([0.4, 1.8, 4.8])
        measures_.table_proportion_variances = table_proportion_variances_
        weighted_bases_.base_values = np.array([10, 20, 30])
        measures_.weighted_bases = weighted_bases_
        table_proportion_stderrs = _TableProportionStderrs(None, measures_, None)

        assert table_proportion_stderrs.base_values == pytest.approx([0.2, 0.3, 0.4])

    def it_computes_its_subtotal_values_to_help(
        self, measures_, table_proportion_variances_, weighted_bases_
    ):
        table_proportion_variances_.subtotal_values = np.array([4.8, 12.5])
        measures_.table_proportion_variances = table_proportion_variances_
        weighted_bases_.subtotal_values = np.array([30, 50])
        measures_.weighted_bases = weighted_bases_
        table_proportion_stderrs = _TableProportionStderrs(None, measures_, None)

        assert table_proportion_stderrs.subtotal_values == pytest.approx([0.4, 0.5])

    # fixture components ---------------------------------------------

    @pytest.fixture
    def measures_(self, request):
        return instance_mock(request, StripeMeasures)

    @pytest.fixture
    def table_proportion_variances_(self, request):
        return instance_mock(request, _TableProportionVariances)

    @pytest.fixture
    def weighted_bases_(self, request):
        return instance_mock(request, _WeightedBases)


class Describe_TableProportionVariances(object):
    """Unit test suite for `cr.cube.stripe.measure._TableProportionVariances` object."""

    def it_computes_its_base_values_to_help(self, measures_, table_proportions_):
        table_proportions_.base_values = np.array([0.2, 0.3, 0.5])
        measures_.table_proportions = table_proportions_
        table_proportion_variances = _TableProportionVariances(None, measures_, None)

        assert table_proportion_variances.base_values == pytest.approx(
            [0.16, 0.21, 0.25]
        )

    def it_computes_its_subtotal_values_to_help(self, measures_, table_proportions_):
        table_proportions_.subtotal_values = np.array([0.5, 0.8])
        measures_.table_proportions = table_proportions_
        table_proportion_variances = _TableProportionVariances(None, measures_, None)

        assert table_proportion_variances.subtotal_values == pytest.approx([0.25, 0.16])

    # fixture components ---------------------------------------------

    @pytest.fixture
    def measures_(self, request):
        return instance_mock(request, StripeMeasures)

    @pytest.fixture
    def table_proportions_(self, request):
        return instance_mock(request, _TableProportions)


class Describe_TableProportions(object):
    """Unit test suite for `cr.cube.stripe.measure._TableProportions` object."""

    @pytest.mark.parametrize(
        "table_margin, expected_value",
        (
            (np.array([4.5, 6.7]), [0.7555556, 0.8358209]),
            (42.42, [0.08015087, 0.1320132]),
        ),
    )
    def it_computes_its_base_values_to_help(
        self,
        measures_,
        weighted_counts_,
        _weighted_cube_counts_prop_,
        weighted_cube_counts_,
        table_margin,
        expected_value,
    ):
        weighted_counts_.base_values = np.array([3.4, 5.6])
        measures_.weighted_counts = weighted_counts_
        weighted_cube_counts_.table_margin = table_margin
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        table_proportions = _TableProportions(None, measures_, None)

        assert table_proportions.base_values == pytest.approx(expected_value)

    @pytest.mark.parametrize(
        "table_margin, expected_value",
        (
            (np.array([4.5, 6.7]), []),
            (42.42, [0.2310231, 0.1791608]),
        ),
    )
    def it_computes_its_subtotal_values_to_help(
        self,
        measures_,
        weighted_counts_,
        _weighted_cube_counts_prop_,
        weighted_cube_counts_,
        table_margin,
        expected_value,
    ):
        weighted_counts_.subtotal_values = np.array([9.8, 7.6])
        measures_.weighted_counts = weighted_counts_
        weighted_cube_counts_.table_margin = table_margin
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        table_proportions = _TableProportions(None, measures_, None)

        assert table_proportions.subtotal_values == pytest.approx(expected_value)

    # fixture components ---------------------------------------------

    @pytest.fixture
    def measures_(self, request):
        return instance_mock(request, StripeMeasures)

    @pytest.fixture
    def weighted_counts_(self, request):
        return instance_mock(request, _WeightedCounts)

    @pytest.fixture
    def weighted_cube_counts_(self, request):
        return instance_mock(request, _BaseWeightedCubeCounts)

    @pytest.fixture
    def _weighted_cube_counts_prop_(self, request):
        return property_mock(request, _TableProportions, "_weighted_cube_counts")


class Describe_UnweightedBases(object):
    """Unit test suite for `cr.cube.stripe.measure._UnweightedBases` object."""

    def it_knows_its_base_values(
        self, _unweighted_cube_counts_prop_, unweighted_cube_counts_
    ):
        _unweighted_cube_counts_prop_.return_value = unweighted_cube_counts_
        unweighted_cube_counts_.bases = np.array([3, 4, 5])
        unweighted_bases = _UnweightedBases(None, None, None)

        assert unweighted_bases.base_values.tolist() == [3, 4, 5]

    @pytest.mark.parametrize(
        "subtotal_values, expected_value",
        ((np.array([]), []), (np.array([5, 3]), [42, 42])),
    )
    def it_knows_its_subtotal_values(
        self, request, _unweighted_cube_counts_prop_, subtotal_values, expected_value
    ):
        rows_dimension_ = instance_mock(request, Dimension)
        property_mock(request, _UnweightedBases, "base_values", return_value=[3, 2, 1])
        SumSubtotals_ = class_mock(request, "cr.cube.stripe.measure.SumSubtotals")
        SumSubtotals_.subtotal_values.return_value = subtotal_values
        _unweighted_cube_counts_prop_.return_value = instance_mock(
            request, _CatUnweightedCubeCounts, table_base=42
        )
        unweighted_bases = _UnweightedBases(rows_dimension_, None, None)

        subtotal_values = unweighted_bases.subtotal_values

        SumSubtotals_.subtotal_values.assert_called_once_with(
            [3, 2, 1], rows_dimension_
        )
        assert subtotal_values.tolist() == expected_value

    def it_knows_its_table_base_range(
        self, _unweighted_cube_counts_prop_, unweighted_cube_counts_
    ):
        _unweighted_cube_counts_prop_.return_value = unweighted_cube_counts_
        unweighted_cube_counts_.bases = np.array([25, 20, 5, 15, 10])
        unweighted_bases = _UnweightedBases(None, None, None)

        assert unweighted_bases.table_base_range.tolist() == [5, 25]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def unweighted_cube_counts_(self, request):
        return instance_mock(request, _BaseUnweightedCubeCounts)

    @pytest.fixture
    def _unweighted_cube_counts_prop_(self, request):
        return property_mock(request, _UnweightedBases, "_unweighted_cube_counts")


class Describe_UnweightedCounts(object):
    """Unit test suite for `cr.cube.stripe.measure._UnweightedCounts` object."""

    def it_knows_its_base_values(
        self, _unweighted_cube_counts_prop_, unweighted_cube_counts_
    ):
        _unweighted_cube_counts_prop_.return_value = unweighted_cube_counts_
        unweighted_cube_counts_.unweighted_counts = np.array([1, 2, 3])
        unweighted_counts = _UnweightedCounts(None, None, None)

        assert unweighted_counts.base_values.tolist() == [1, 2, 3]

    def it_knows_its_subtotal_values(self, request):
        rows_dimension_ = instance_mock(request, Dimension)
        property_mock(request, _UnweightedCounts, "base_values", return_value=[1, 2, 3])
        SumSubtotals_ = class_mock(request, "cr.cube.stripe.measure.SumSubtotals")
        SumSubtotals_.subtotal_values.return_value = np.array([3, 5])
        unweighted_counts = _UnweightedCounts(rows_dimension_, None, None)

        subtotal_values = unweighted_counts.subtotal_values

        SumSubtotals_.subtotal_values.assert_called_once_with(
            [1, 2, 3], rows_dimension_
        )
        assert subtotal_values.tolist() == [3, 5]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def unweighted_cube_counts_(self, request):
        return instance_mock(request, _BaseUnweightedCubeCounts)

    @pytest.fixture
    def _unweighted_cube_counts_prop_(self, request):
        return property_mock(request, _UnweightedCounts, "_unweighted_cube_counts")


class Describe_WeightedBases(object):
    """Unit test suite for `cr.cube.stripe.measure._WeightedBases` object."""

    def it_knows_its_base_values(
        self, _weighted_cube_counts_prop_, weighted_cube_counts_
    ):
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        weighted_cube_counts_.bases = np.array([3.3, 4.4, 5.5])
        weighted_bases = _WeightedBases(None, None, None)

        assert weighted_bases.base_values.tolist() == [3.3, 4.4, 5.5]

    @pytest.mark.parametrize(
        "subtotal_values, expected_value",
        ((np.array([]), []), (np.array([5.5, 3.3]), [42.24, 42.24])),
    )
    def it_knows_its_subtotal_values(
        self,
        request,
        _weighted_cube_counts_prop_,
        subtotal_values,
        expected_value,
    ):
        rows_dimension_ = instance_mock(request, Dimension)
        property_mock(
            request, _WeightedBases, "base_values", return_value=[3.3, 2.2, 1.1]
        )
        SumSubtotals_ = class_mock(request, "cr.cube.stripe.measure.SumSubtotals")
        SumSubtotals_.subtotal_values.return_value = subtotal_values
        _weighted_cube_counts_prop_.return_value = instance_mock(
            request, _CatWeightedCubeCounts, table_margin=42.24
        )
        weighted_bases = _WeightedBases(rows_dimension_, None, None)

        subtotal_values = weighted_bases.subtotal_values

        SumSubtotals_.subtotal_values.assert_called_once_with(
            [3.3, 2.2, 1.1], rows_dimension_
        )
        assert subtotal_values.tolist() == expected_value

    def it_knows_its_table_margin_range(
        self, _weighted_cube_counts_prop_, weighted_cube_counts_
    ):
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        weighted_cube_counts_.bases = np.array([25.5, 20.2, 5.5, 15.2, 10.0])
        weighted_bases = _WeightedBases(None, None, None)

        assert weighted_bases.table_margin_range.tolist() == [5.5, 25.5]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def weighted_cube_counts_(self, request):
        return instance_mock(request, _BaseWeightedCubeCounts)

    @pytest.fixture
    def _weighted_cube_counts_prop_(self, request):
        return property_mock(request, _WeightedBases, "_weighted_cube_counts")


class Describe_WeightedCounts(object):
    """Unit test suite for `cr.cube.stripe.measure._WeightedCounts` object."""

    def it_computes_its_base_values_to_help(
        self, _weighted_cube_counts_prop_, weighted_cube_counts_
    ):
        _weighted_cube_counts_prop_.return_value = weighted_cube_counts_
        weighted_cube_counts_.weighted_counts = np.array([1, 2, 3])
        weighted_counts = _WeightedCounts(None, None, None)

        assert weighted_counts.base_values.tolist() == [1, 2, 3]

    def it_computes_its_subtotal_values_to_help(self, request):
        rows_dimension_ = instance_mock(request, Dimension)
        property_mock(
            request, _WeightedCounts, "base_values", return_value=[1.1, 2.2, 3.3]
        )
        SumSubtotals_ = class_mock(request, "cr.cube.stripe.measure.SumSubtotals")
        SumSubtotals_.subtotal_values.return_value = np.array([3.3, 5.5])
        weighted_counts = _WeightedCounts(rows_dimension_, None, None)

        subtotal_values = weighted_counts.subtotal_values

        SumSubtotals_.subtotal_values.assert_called_once_with(
            [1.1, 2.2, 3.3], rows_dimension_
        )
        assert subtotal_values.tolist() == [3.3, 5.5]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def weighted_cube_counts_(self, request):
        return instance_mock(request, _BaseWeightedCubeCounts)

    @pytest.fixture
    def _weighted_cube_counts_prop_(self, request):
        return property_mock(request, _WeightedCounts, "_weighted_cube_counts")



# encoding: utf-8

"""Provides the CubeSet and Cube classes.

CubeSet is the main API class for manipulating Crunch.io JSON cube responses.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import json

import numpy as np

from cr.cube.cubepart import CubePartition
from cr.cube.dimension import AllDimensions
from cr.cube.enums import CUBE_MEASURE, DIMENSION_TYPE as DT, NUMERIC_MEASURES
from cr.cube.util import lazyproperty

np.seterr(divide="ignore", invalid="ignore")


class CubeSet(object):
    """Represents a multi-cube cube-response.

    Also works just fine for a single cube-response passed inside a sequence, allowing
    uniform handling of single and multi-cube responses.

    `cube_responses` is a sequence of cube-response dicts received from Crunch. The
    sequence can contain a single item, such as a cube-response for a slide, but it must
    be contained in a sequence. A tabbook cube-response sequence can be passed as it was
    received.

    `transforms` is a sequence of transforms dicts corresponding in order to the
    cube-responses. `population` is the estimated target population and is used when
    a population-projection measure is requested. `min_base` is an integer representing
    the minimum sample-size used for indicating values that are unreliable by reason of
    insufficient sample (base).
    """

    def __init__(self, cube_responses, transforms, population, min_base):
        self._cube_responses = cube_responses
        self._transforms_dicts = transforms
        self._population = population
        self._min_base = min_base

    @lazyproperty
    def available_measures(self):
        """frozenset of available measures of the first cube in this set."""
        return self._cubes[0].available_measures

    @lazyproperty
    def can_show_pairwise(self):
        """True if all 2D cubes in a multi-cube set can provide pairwise comparison."""
        if len(self._cubes) < 2:
            return False

        return all(
            all(dt in DT.ALLOWED_PAIRWISE_TYPES for dt in cube.dimension_types[-2:])
            and cube.ndim >= 2
            for cube in self._cubes[1:]
        )

    @lazyproperty
    def description(self):
        """str description of first cube in this set."""
        return self._cubes[0].description

    @lazyproperty
    def has_weighted_counts(self):
        """True if cube-responses include a weighted-count measure."""
        return self._cubes[0].has_weighted_counts

    @lazyproperty
    def is_ca_as_0th(self):
        """True for multi-cube when first cube represents a categorical-array.

        A "CA-as-0th" tabbook tab is "3D" in the sense it is "sliced" into one table
        (partition-set) for each of the CA subvariables.
        """
        # ---can only be true for multi-cube case---
        if not self._is_multi_cube:
            return False
        # ---the rest depends on the row-var cube---
        cube = self._cubes[0]
        # ---True if row-var cube is CA---
        return cube.dimension_types[0] == DT.CA_SUBVAR

    @lazyproperty
    def missing_count(self):
        """The number of missing values from first cube in this set."""
        return self._cubes[0].missing

    @lazyproperty
    def name(self):
        """str name of first cube in this set."""
        return self._cubes[0].name

    @lazyproperty
    def partition_sets(self):
        """Sequence of cube-partition collections across all cubes of this cube-set.

        This value might look like the following for a ca-as-0th tabbook, for example:

            (
                (_Strand, _Slice, _Slice),
                (_Strand, _Slice, _Slice),
                (_Strand, _Slice, _Slice),
            )

        and might often look like this for a typical slide:

            ((_Slice,))

        Each partition set represents the partitions for a single "stacked" table. A 2D
        slide has a single partition-set of a single _Slice object, as in the second
        example above. A 3D slide would have multiple partition sets, each of a single
        _Slice. A tabook will have multiple partitions in each set, the first being
        a _Strand and the rest being _Slice objects. Multiple partition sets only arise
        for a tabbook in the CA-as-0th case.
        """
        return tuple(zip(*(cube.partitions for cube in self._cubes)))

    @lazyproperty
    def population_fraction(self):
        """The filtered/unfiltered ratio for this cube-set.

        This value is required for properly calculating population on a cube where
        a filter has been applied. Returns 1.0 for an unfiltered cube. Returns `np.nan`
        if the unfiltered count is zero, which would otherwise result in
        a divide-by-zero error.
        """
        return self._cubes[0].population_fraction

    @lazyproperty
    def n_responses(self):
        """Total number of responses considered from first cube in this set."""
        return self._cubes[0].n_responses

    @lazyproperty
    def valid_counts_summary(self):
        """The valid count summary values from first cube in this set."""
        return self._cubes[0].valid_counts_summary

    @lazyproperty
    def _cubes(self):
        """Sequence of Cube objects containing data for this analysis."""

        def iter_cubes():
            """Generate a Cube object for each of cube_responses.

            0D cube-responses and 1D second-and-later cubes are "inflated" to add their
            missing row dimension.
            """
            for idx, cube_response in enumerate(self._cube_responses):
                cube = Cube(
                    cube_response,
                    cube_idx=idx if self._is_multi_cube else None,
                    transforms=self._transforms_dicts[idx],
                    population=self._population,
                    mask_size=self._min_base,
                )
                # --- numeric-measures cubes require inflation to restore their
                # --- rows-dimension, others don't
                yield cube.inflate() if self._is_numeric_measure else cube

        return tuple(iter_cubes())

    @lazyproperty
    def _is_multi_cube(self):
        """True if more than one cube-response was provided on construction."""
        return len(self._cube_responses) > 1

    @lazyproperty
    def _is_numeric_measure(self):
        """True when CubeSet is special-case "numeric-measure" case requiring inflation.

        When a numeric variable with `mean`, `sum` or `std_dev` summary statistic
        expressed in its view, appears as the rows-dimension in a multitable analysis,
        its cube-result has been "reduced" to the mean-value of those numerics. This is
        in contrast to being "bucketized" into an arbitrary set of numeric-range
        categories like 0-5, 5-10, etc. In the process, as an artifact of the ZZ9 query
        response, that dimension is removed. As a result, the rows-dimension cube is 0D
        and the column-dimension cubes are 1D. These need to be "inflated" to restore
        the lost dimension such that they are uniform with other cube-results and can be
        processed without special-case code.

        "Inflation" is basically prefixing "1 x" to the dimensionality, for example a 1D
        of size 5 becomes a 1 x 5 2D result. Note this requires no mapping in the actual
        values because 5 = 1 x 5 = 5 (values).
        """
        # --- this case only arises in a multitable analysis ---
        if not self._is_multi_cube:
            return False

        # --- We need the cube to tell us the dimensionality. This redundant
        # --- construction is low-overhead because all Cube properties are lazy.
        return Cube(self._cube_responses[0]).ndim == 0


class Cube(object):
    """Provides access to individual slices on a cube-result.

    It also provides some attributes of the overall cube-result.

    `cube_idx` must be `None` (or omitted) for a single-cube CubeSet. This indicates the
    CubeSet contains only a single cube and influences behaviors like CA-as-0th.
    """

    def __init__(
        self, response, cube_idx=None, transforms=None, population=None, mask_size=0
    ):
        self._cube_response_arg = response
        self._transforms_dict = {} if transforms is None else transforms
        self._cube_idx_arg = cube_idx
        self._population = 0 if population is None else population
        self._mask_size = mask_size

    def __repr__(self):
        """Provide text representation suitable for working at console.

        Falls back to a default repr on exception, such as might occur in
        unit tests where object need not otherwise be provided with all
        instance variable values.
        """
        try:
            dimensionality = " x ".join(dt.name for dt in self.dimension_types)
            return "%s(name='%s', dimension_types='%s')" % (
                type(self).__name__,
                self.name,
                dimensionality,
            )
        except Exception:
            return super(Cube, self).__repr__()

    @lazyproperty
    def available_measures(self):
        """frozenset of available CUBE_MEASURE members in the cube response."""
        cube_measures = self._cube_response.get("result", {}).get("measures", {}).keys()
        return frozenset(CUBE_MEASURE(m) for m in cube_measures)

    @lazyproperty
    def counts(self):
        return self.counts_with_missings[self._valid_idxs]

    @lazyproperty
    def counts_with_missings(self):
        """ndarray of weighted, unweighted or valid counts including missing values.

        The difference from .counts is that this property includes value for missing
        categories.
        """
        return (
            self._measures.valid_counts.raw_cube_array
            if self._measures.valid_counts is not None
            else self._measures.weighted_counts.raw_cube_array
            if self.has_weighted_counts
            else self._measures.unweighted_counts.raw_cube_array
        )

    @lazyproperty
    def covariance(self):
        """Optional float64 ndarray of the cube_covariance if the measure exists."""
        if self._measures.covariance is None:
            return None
        return self._measures.covariance.raw_cube_array[self._valid_idxs].astype(
            np.float64
        )

    @lazyproperty
    def cube_index(self):
        """Offset of this cube within its CubeSet."""
        return 0 if self._cube_idx_arg is None else self._cube_idx_arg

    @lazyproperty
    def description(self):
        """Return the description of the cube."""
        if not self.dimensions:
            return None
        return self.dimensions[0].description

    @lazyproperty
    def dimension_types(self):
        """Tuple of DIMENSION_TYPE member for each dimension of cube."""
        return tuple(d.dimension_type for d in self.dimensions)

    @lazyproperty
    def dimensions(self):
        """_ApparentDimensions object providing access to visible dimensions.

        A cube involving a multiple-response (MR) variable has two dimensions
        for that variable (subvariables and categories dimensions), but is
        "collapsed" into a single effective dimension for cube-user purposes
        (its categories dimension is supressed). This collection will contain
        a single dimension for each MR variable and therefore may have fewer
        dimensions than appear in the cube response.
        """
        return self._all_dimensions.apparent_dimensions

    def inflate(self):
        """Return new Cube object with rows-dimension added.

        A multi-cube (tabbook) response formed from a function (e.g. mean()) on
        a numeric variable arrives without a rows-dimension.
        """
        cube_dict = self._cube_dict
        dimensions = cube_dict["result"]["dimensions"]
        default_name = "-".join([m.value for m in self._available_numeric_measures])
        # --- The default value in case of numeric variable is the combination of all
        # --- the measures expressed in the cube response.
        alias = self._numeric_measure_references.get("alias", default_name)
        name = self._numeric_measure_references.get("name", default_name).title()
        rows_dimension = {
            "references": {"alias": alias, "name": name},
            "type": {
                "categories": [{"id": 1, "name": name}],
                "class": "categorical",
            },
        }
        dimensions.insert(0, rows_dimension)
        return Cube(
            cube_dict,
            self._cube_idx_arg,
            self._transforms_dict,
            self._population,
            self._mask_size,
        )

    @lazyproperty
    def has_weighted_counts(self):
        """True if cube response has weighted count data."""
        return self.weighted_counts is not None

    @lazyproperty
    def means(self):
        """Optional float64 ndarray of the cube_means if the measure exists."""
        if self._measures.means is None:
            return None
        return self._measures.means.raw_cube_array[self._valid_idxs].astype(np.float64)

    @lazyproperty
    def missing(self):
        """Get missing count of a cube."""
        return self._measures.missing_count

    @lazyproperty
    def name(self):
        """Return the name of the cube.

        If the cube has 2 diensions, return the name of the second one. In case
        of a different number of dimensions, default to returning the name of
        the last one. In case of no dimensions, return the empty string.
        """
        if not self.dimensions:
            return None
        return self.dimensions[0].name

    @lazyproperty
    def ndim(self):
        """int count of dimensions for this cube."""
        return len(self.dimensions)

    @lazyproperty
    def n_responses(self):
        """Total (int) number of responses considered."""
        return self._cube_response["result"].get("n", 0)

    @lazyproperty
    def overlaps(self):
        """Optional float64 ndarray of cube_overlaps if the measure exists.

        The array has as many dimensions as there are defined in the cube query, plus
        the extra subvariables dimension as the last dimension.
        """
        if self._measures.overlaps is None:
            return None
        return self._measures.overlaps.raw_cube_array[self._valid_idxs].astype(
            np.float64
        )

    @lazyproperty
    def partitions(self):
        """Sequence of _Slice, _Strand, or _Nub objects from this cube-result."""
        return tuple(
            CubePartition.factory(
                self,
                slice_idx=slice_idx,
                transforms=self._transforms_dict,
                population=self._population,
                ca_as_0th=self._ca_as_0th,
                mask_size=self._mask_size,
            )
            for slice_idx in self._slice_idxs
        )

    @lazyproperty
    def population_fraction(self):
        """The filtered/unfiltered ratio for cube response.

        This value is required for properly calculating population on a cube
        where a filter has been applied. Returns 1.0 for an unfiltered cube.
        Returns `np.nan` if the unfiltered count is zero, which would
        otherwise result in a divide-by-zero error.
        """
        return self._measures.population_fraction

    @lazyproperty
    def stddev(self):
        """Optional float64 ndarray of the cube_stddev if the measure exists."""
        if self._measures.stddev is None:
            return None
        return self._measures.stddev.raw_cube_array[self._valid_idxs].astype(np.float64)

    @lazyproperty
    def sums(self):
        """Optional float64 ndarray of the cube_sum if the measure exists."""
        if self._measures.sums is None:
            return None
        return self._measures.sums.raw_cube_array[self._valid_idxs].astype(np.float64)

    @lazyproperty
    def title(self):
        """str alternate-name given to cube-result.

        This value is suitable for naming a Strand when displayed as a column. In this
        use-case it is a stand-in for the columns-dimension name since a strand has no
        columns dimension.
        """
        return self._cube_dict["result"].get("title", "Untitled")

    @lazyproperty
    def unweighted_counts(self):
        """ndarray of unweighted counts, valid elements only.

        Unweighted counts are drawn from the `result.counts` field of the cube result.
        These counts are always present, even when the measure is numeric and there are
        no count measures. These counts are always unweighted, regardless of whether the
        cube is "weighted".

        In case of presence of valid counts in the cube response the counts are replaced
        with the valid counts measure.
        """
        unweighted_counts = (
            self._measures.valid_counts
            if self._measures.valid_counts is not None
            else self._measures.unweighted_counts
        )
        return unweighted_counts.raw_cube_array[self._valid_idxs]

    @lazyproperty
    def valid_counts_summary(self):
        """Optional ndarray of summary valid counts"""
        if not self._measures.valid_counts:
            return None
        # --- In case of ndim >= 2 the sum should be done on the second axes to get
        # --- the correct sequence of valid count (e.g. CA_SUBVAR).
        axis = 1 if len(self._all_dimensions) >= 2 else 0
        return np.sum(
            self._measures.valid_counts.raw_cube_array[self._valid_idxs], axis=axis
        )

    @lazyproperty
    def valid_overlaps(self):
        """Optional float64 ndarray of cube_valid_overlaps if the measure exists.

        The array has as many dimensions as there are defined in the cube query, plus
        the extra subvariables dimension as the last dimension.
        """
        if self._measures.valid_overlaps is None:
            return None  # pragma: no cover
        return self._measures.valid_overlaps.raw_cube_array[self._valid_idxs].astype(
            np.float64
        )

    @lazyproperty
    def weighted_counts(self):
        """ndarray of weighted counts, valid elements only.

        In case of presence of valid counts in the cube response the weighted counts
        are replaced with the valid counts measure.
        """
        weighted_counts = (
            self._measures.valid_counts
            if self._measures.valid_counts is not None
            else self._measures.weighted_counts
        )
        return (
            weighted_counts.raw_cube_array[self._valid_idxs]
            if weighted_counts is not None
            else None
        )

    @lazyproperty
    def _all_dimensions(self):
        """The AllDimensions object for this cube.

        The AllDimensions object provides access to all the dimensions appearing in the
        cube response, not only apparent dimensions (those that appear to a user). It
        also provides access to an _ApparentDimensions object which contains only those
        user-apparent dimensions (basically the categories dimension of each MR
        dimension-pair is suppressed).
        """
        return AllDimensions(dimension_dicts=self._cube_dict["result"]["dimensions"])

    @lazyproperty
    def _available_numeric_measures(self):
        """tuple of available numeric measures expressed in the cube_response.

        Basically the numeric measures are the intersection between all the measures
        within the cube response and the defined NUMERIC_MEASURES.
        """
        return tuple(self.available_measures.intersection(NUMERIC_MEASURES))

    @lazyproperty
    def _ca_as_0th(self):
        """True if slicing is to be performed in so-called "CA-as-0th" mode.

        In this mode, a categorical-array (CA) cube (2D) is sliced into a sequence of 1D
        slices, each of which represents one subvariable of the CA variable. Normally,
        a 2D cube-result becomes a single slice.
        """
        return (
            (self._cube_idx_arg == 0 or self._is_single_filter_col_cube)
            and len(self.dimension_types) > 0
            and self.dimension_types[0] == DT.CA
        )

    @lazyproperty
    def _cube_dict(self):
        """dict containing raw cube response, parsed from JSON payload."""
        cube_dict = copy.deepcopy(self._cube_response)
        if self._numeric_measure_subvariables:
            dimensions = cube_dict.get("result", {}).get("dimensions", [])
            # ---dim inflation---
            # ---In case of numeric arrays, we need to inflate the row dimension
            # ---according to the mean subvariables. For each subvar the row dimension
            # ---will have a new element related to the subvar metadata.
            dimensions.insert(0, self._numeric_array_dimension)
        return cube_dict

    @lazyproperty
    def _cube_response(self):
        """dict representing the parsed cube response arguments."""
        try:
            response = self._cube_response_arg
            # ---parse JSON to a dict when constructed with JSON---
            cube_response = (
                response if isinstance(response, dict) else json.loads(response)
            )
            # ---cube is 'value' item in a shoji response---
            return cube_response.get("value", cube_response)
        except TypeError:
            raise TypeError(
                "Unsupported type <%s> provided. Cube response must be JSON "
                "(str) or dict." % type(self._cube_response_arg).__name__
            )

    @lazyproperty
    def _is_single_filter_col_cube(self):
        """bool determines if it is a single column filter cube."""
        return self._cube_dict["result"].get("is_single_col_cube", False)

    @lazyproperty
    def _measures(self):
        """_Measures object for this cube.

        Provides access to count based measures and numeric measures (e.g. mean, sum)
        when available.
        """
        return _Measures(self._cube_dict, self._all_dimensions, self._cube_idx_arg)

    @lazyproperty
    def _numeric_measure_references(self):
        """Dict of numeric measure references, typically for numeric measures."""
        if not self._available_numeric_measures:
            return {}
        cube_response = self._cube_response
        cube_measures = cube_response.get("result", {}).get("measures", {})
        metadata = cube_measures.get(self._available_numeric_measures[0].value, {}).get(
            "metadata", {}
        )
        return metadata.get("references", {})

    @lazyproperty
    def _numeric_measure_subvariables(self):
        """List of mean subvariables, typically for numeric arrays."""
        if not self._available_numeric_measures:
            return []
        cube_response = self._cube_response
        cube_measures = cube_response.get("result", {}).get("measures", {})
        metadata = cube_measures.get(self._available_numeric_measures[0].value, {}).get(
            "metadata", {}
        )
        return metadata.get("type", {}).get("subvariables", [])

    @lazyproperty
    def _numeric_array_dimension(self):
        """Rows dimension object according to the numeric-measure subvariables."""
        if not self._numeric_measure_subvariables:
            return None
        subrefs = self._numeric_measure_references.get("subreferences", [])
        rows_dimension = {
            "references": {
                "alias": self._numeric_measure_references.get("alias"),
                "name": self._numeric_measure_references.get("name"),
            },
            "type": {"elements": [], "class": "enum", "subtype": {"class": "num_arr"}},
        }
        # ---In case of numeric arrays the row dimension should contains additional
        # ---information related to the subreferences for each subvariable of the
        # ---array.
        for i, _ in enumerate(self._numeric_measure_subvariables):
            # ---The row dimensions elements must be expanded with the alias and the
            # ---name of the numeric array mean measure subreferences.
            rows_dimension["type"].get("elements", []).append(
                {
                    "id": i,
                    "value": {
                        "references": {
                            "alias": subrefs[i].get("alias") if subrefs else None,
                            "name": subrefs[i].get("name") if subrefs else None,
                        },
                        "id": self._numeric_measure_subvariables[i],
                    },
                },
            )
        return rows_dimension

    @lazyproperty
    def _slice_idxs(self):
        """Iterable of contiguous int indicies for slices to be produced.

        This value is to help cube-section construction which does not by itself know
        how many slices are in a cube-result.
        """
        if self.ndim < 3 and not self._ca_as_0th:
            return (0,)
        return range(len(self.dimensions[0].valid_elements))

    @lazyproperty
    def _valid_idxs(self):
        """Tuple of int64 ndarrays of the valid elements idx for each dimension."""
        valid_idxs = np.ix_(
            *tuple(d.valid_elements.element_idxs for d in self._all_dimensions)
        )
        # The dimension dimension order can change in case of numeric array variable on
        # the row, and so valid indices needs to be returned in an ordered way.
        return tuple(valid_idxs[i] for i in self._all_dimensions.dimension_order)


class _Measures(object):
    """Provides access to measures contained in cube response."""

    def __init__(self, cube_dict, all_dimensions, cube_idx_arg=None):
        self._cube_dict = cube_dict
        self._all_dimensions = all_dimensions
        self._cube_idx_arg = cube_idx_arg

    @lazyproperty
    def covariance(self):
        """Optional _CovarianceMeasure object providing access to covariance values.

        Will be None if covariance is not available int the cube response.
        """
        covariance = _CovarianceMeasure(
            self._cube_dict, self._all_dimensions, self._cube_idx_arg
        )
        return None if covariance.raw_cube_array is None else covariance

    @lazyproperty
    def means(self):
        """Optional _MeanMeasure object providing access to means values.

        Will be None if no means are available on the counts.
        """
        mean = _MeanMeasure(self._cube_dict, self._all_dimensions, self._cube_idx_arg)
        return None if mean.raw_cube_array is None else mean

    @lazyproperty
    def missing_count(self):
        """numeric representing count of missing rows in cube response."""
        if self.valid_counts is not None:
            return self.valid_counts.missing_count
        # The check on the means measure is needed for retro-compatibility with the old
        # fixtures that don't have valid_counts.
        if self.means is not None:
            return self.means.missing_count
        return self._cube_dict["result"].get("missing", 0)

    @lazyproperty
    def overlaps(self):
        """Optional _OverlapMeasure object providing access to overlaps values.

        Will be None if no overlaps are available on the cube result.
        """
        overlap = _OverlapMeasure(
            self._cube_dict, self._all_dimensions, self._cube_idx_arg
        )
        return None if overlap.raw_cube_array is None else overlap

    @lazyproperty
    def population_fraction(self):
        """The filtered/unfiltered ratio for cube response.

        The filtered counts are calculated for complete-cases. This means that only the
        non-missing entries are included in the filtered counts. Complete cases are
        used only if the corresponding cases are included in the cube response. If not,
        the old-style default calculation is used.

        This value is required for properly calculating population on a cube
        where a filter has been applied. Returns 1.0 for an unfiltered cube.
        Returns `np.nan` if the unfiltered count is zero, which would
        otherwise result in a divide-by-zero error.
        """

        # Try and get the new-style complete-cases filtered counts
        filter_stats = (
            self._cube_dict["result"]
            .get("filter_stats", {})
            .get("filtered_complete", {})
            .get("weighted")
        )
        if filter_stats:
            # If new format is present in response json, use that for pop fraction
            numerator = filter_stats["selected"]
            denominator = numerator + filter_stats["other"]
        else:
            # If new format is not available, default to old-style calculation
            numerator = self._cube_dict["result"].get("filtered", {}).get("weighted_n")
            denominator = (
                self._cube_dict["result"].get("unfiltered", {}).get("weighted_n")
            )

        try:
            return numerator / denominator
        except ZeroDivisionError:
            return np.nan
        except Exception:
            return 1.0

    @lazyproperty
    def stddev(self):
        """_StdDevMeasure object providing access to cube stddev values.

        None when the cube response does not contain a stddev measure.
        """
        stddev = _StdDevMeasure(
            self._cube_dict, self._all_dimensions, self._cube_idx_arg
        )
        return None if stddev.raw_cube_array is None else stddev

    @lazyproperty
    def sums(self):
        """_SumMeasure object providing access to cube sum values.

        None when the cube response does not contain a sum measure.
        """
        sums = _SumMeasure(self._cube_dict, self._all_dimensions, self._cube_idx_arg)
        return None if sums.raw_cube_array is None else sums

    @lazyproperty
    def unweighted_counts(self):
        """_UnweightedCountMeasure object for this cube.

        This object provides access to unweighted counts for this cube,
        whether or not the cube contains weighted counts.
        """
        return _UnweightedCountMeasure(
            self._cube_dict, self._all_dimensions, self._cube_idx_arg
        )

    @lazyproperty
    def valid_counts(self):
        """_ValidCountsMeasure object for this cube.

        Can be None when cube doesn't have valid counts.
        """
        valid_counts = _ValidCountsMeasure(
            self._cube_dict, self._all_dimensions, self._cube_idx_arg
        )
        return valid_counts if valid_counts.raw_cube_array is not None else None

    @lazyproperty
    def valid_overlaps(self):
        """Optional _ValidOverlapMeasure object providing access to valid overlaps vals.

        Will be None if no valid overlaps are available on the cube result.
        """
        overlap = _ValidOverlapMeasure(
            self._cube_dict, self._all_dimensions, self._cube_idx_arg
        )
        return None if overlap.raw_cube_array is None else overlap

    @lazyproperty
    def weighted_counts(self):
        """Optional _WeightedCountMeasure object for this cube.

        Can be None when the cube is unweighted.
        """
        weighted_counts = _WeightedCountMeasure(
            self._cube_dict, self._all_dimensions, self._cube_idx_arg
        )
        return weighted_counts if weighted_counts.raw_cube_array is not None else None


class _BaseMeasure(object):
    """Base class for measure objects."""

    def __init__(self, cube_dict, all_dimensions, cube_idx_arg=None):
        self._cube_dict = cube_dict
        self._all_dimensions = all_dimensions
        self._cube_idx_arg = cube_idx_arg

    @lazyproperty
    def raw_cube_array(self):
        """Optional read-only ndarray of measure values from cube-response.

        The shape of the ndarray mirrors the shape of the (raw) cube
        response. Specifically, it includes values for missing elements, any
        MR_CAT dimensions, and any prunable rows and columns. Returns None
        if the measure is not available in cube.
        """
        if self._flat_values is None:
            return None
        raw_cube_array = self._flat_values.reshape(self._shape)
        # ---must be read-only to avoid hard-to-find bugs---
        raw_cube_array.flags.writeable = False
        return raw_cube_array

    @lazyproperty
    def _flat_values(self):  # pragma: no cover
        """Return ndarray of np.float64 values as found in cube response.

        This property must be implemented by each subclass.
        """
        raise NotImplementedError("must be implemented by each subclass")

    @lazyproperty
    def _shape(self):
        """tuple(int) representing the shape of the raw-cube measure array.

        If needed, this property can be overridden, to accustom different measure shapes
        even if the basic cube has the same original shape.
        """
        return self._all_dimensions.shape


class _CovarianceMeasure(_BaseMeasure):
    """Covariance values from a cube-response."""

    @lazyproperty
    def _flat_values(self):
        """Optional 1D np.ndarray of np.float64 cov values as found in cube response.

        Covariance data may include missing items represented by a dict like
        {'?': -1} in the cube response. These are replaced by np.nan in the
        returned value.
        """
        if self._measure_payload is None:
            return None
        return np.array(
            tuple(
                np.nan if type(x) is dict else x for x in self._measure_payload["data"]
            ),
            dtype=np.float64,
        ).flatten()

    @lazyproperty
    def _measure_payload(self):
        """dict representing the covariance measure part of the cube response."""
        return self._cube_dict["result"].get("measures", {}).get("covariance")

    @lazyproperty
    def _numeric_measure_subvariables(self):
        """List of subvariables, typically for numeric arrays."""
        metadata = self._measure_payload.get("metadata", {})
        return metadata.get("type", {}).get("subvariables", [])

    @lazyproperty
    def _shape(self):
        """tuple(int) representing the shape of the covariance."""
        return self._all_dimensions.shape + (len(self._numeric_measure_subvariables),)


class _MeanMeasure(_BaseMeasure):
    """Statistical mean values from a cube-response."""

    @lazyproperty
    def missing_count(self):
        """Numeric value representing count of missing rows in response."""
        return self._cube_dict["result"]["measures"]["mean"].get("n_missing", 0)

    @lazyproperty
    def _flat_values(self):
        """Optional 1D np.ndarray of np.float64 mean values as found in cube response.

        Mean data may include missing items represented by a dict like
        {'?': -1} in the cube response. These are replaced by np.nan in the
        returned value.
        """
        measure_payload = self._cube_dict["result"].get("measures", {}).get("mean")
        if measure_payload is None:
            return None
        return np.array(
            tuple(np.nan if type(x) is dict else x for x in measure_payload["data"]),
            dtype=np.float64,
        ).flatten()


class _OverlapMeasure(_BaseMeasure):
    """Overlap values from a cube-response."""

    @lazyproperty
    def _flat_values(self):
        """Optional 1D np.ndarray of np.float64 overlap values as found in cube response.

        Overlap data may include missing items represented by a dict like
        {'?': -1} in the cube response. These are replaced by np.nan in the
        returned value.
        """
        if self._measure_payload is None:
            return None
        return np.array(
            tuple(
                np.nan if type(x) is dict else x for x in self._measure_payload["data"]
            ),
            dtype=np.float64,
        ).flatten()

    @lazyproperty
    def _measure_payload(self):
        """dict representing the overlaps measure part of the cube response."""
        return self._cube_dict["result"].get("measures", {}).get("overlap")

    @lazyproperty
    def _shape(self):
        """tuple(int) representing shape of the overlaps measure.

        The overlaps measure is characteristic in that it produces an additional
        dimension for the Multiple Response subvariables. That dimension is always
        found at the end of the result shape, because of how responses are
        generated by the backend (ZZ9) mechanism.
        """
        n_subvars = len(self._measure_payload["metadata"]["type"]["subvariables"])
        return self._all_dimensions.shape + (n_subvars,)


class _StdDevMeasure(_BaseMeasure):
    """Statistical stddev values from a cube-response."""

    @lazyproperty
    def _flat_values(self):
        """Optional 1D float64 ndarray of stddev values as found in cube response.

        StdDev data may include missing items represented by a dict like
        {'?': -1} in the cube response. These are replaced by np.nan in the
        returned value.
        """
        measure_payload = self._cube_dict["result"].get("measures", {}).get("stddev")
        if measure_payload is None:
            return None

        return np.array(
            tuple(np.nan if type(x) is dict else x for x in measure_payload["data"]),
            dtype=np.float64,
        ).flatten()


class _SumMeasure(_BaseMeasure):
    """Statistical sum values from a cube-response."""

    @lazyproperty
    def _flat_values(self):
        """Optional 1D float64 ndarray of sum values as found in cube response.

        Sum data may include missing items represented by a dict like
        {'?': -1} in the cube response. These are replaced by np.nan in the
        returned value.
        """
        measure_payload = self._cube_dict["result"].get("measures", {}).get("sum")
        if measure_payload is None:
            return None

        return np.array(
            tuple(np.nan if type(x) is dict else x for x in measure_payload["data"]),
            dtype=np.float64,
        ).flatten()


class _UnweightedCountMeasure(_BaseMeasure):
    """Unweighted counts for cube."""

    @lazyproperty
    def _flat_values(self):
        """1D np.ndarray of np.float64 counts before weighting.

        Use np.float64s to avoid int overflow bugs and so we can use nan.
        """
        return np.array(self._cube_dict["result"]["counts"], dtype=np.float64)


class _ValidCountsMeasure(_BaseMeasure):
    """Valid counts for cube."""

    @lazyproperty
    def missing_count(self):
        """numeric representing count of missing rows reflected in response."""
        return self._cube_dict["result"]["measures"]["valid_count_unweighted"].get(
            "n_missing", 0
        )

    @lazyproperty
    def _flat_values(self):
        """Optional 1D np.ndarray of np.float64 valid counts."""
        valid_counts = (
            self._cube_dict["result"]["measures"]
            .get("valid_count_unweighted", {})
            .get("data", [])
        )
        return np.array(valid_counts, dtype=np.float64) if valid_counts else None


class _ValidOverlapMeasure(_OverlapMeasure):
    """Valid overlap values from a cube-response."""

    @lazyproperty
    def _measure_payload(self):
        """dict representing the valid overlaps measure part of the cube response."""
        return self._cube_dict["result"].get("measures", {}).get("valid_overlap")


class _WeightedCountMeasure(_BaseMeasure):
    """Weighted counts for cube."""

    @lazyproperty
    def _flat_values(self):
        """Optional 1D np.ndarray of np.float64 numeric counts after weighting."""
        unweighted_counts = self._cube_dict["result"]["counts"]
        weighted_counts = (
            self._cube_dict["result"]["measures"].get("count", {}).get("data")
        )
        if unweighted_counts == weighted_counts or weighted_counts is None:
            return None

        return np.array(weighted_counts, dtype=np.float64)



# encoding: utf-8

"""Provides abstracted cube-measure objects used as the basis for second-order measures.

There are several cube-measures that can appear in a cube-response, including
unweighted-counts, weighted-counts (aka. counts), means, and others.
"""

from __future__ import division

import numpy as np

from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.util import lazyproperty


class CubeMeasures(object):
    """Provides access to all cube-measure objects for this stripe."""

    def __init__(self, cube, rows_dimension, ca_as_0th, slice_idx):
        self._cube = cube
        self._rows_dimension = rows_dimension
        self._ca_as_0th = ca_as_0th
        self._slice_idx = slice_idx

    @lazyproperty
    def cube_means(self):
        """_BaseCubeMeans subclass object for this stripe."""
        return _BaseCubeMeans.factory(self._cube, self._rows_dimension)

    @lazyproperty
    def cube_stddev(self):
        """_BaseCubeStdDev subclass object for this stripe."""
        return _BaseCubeStdDev.factory(self._cube, self._rows_dimension)

    @lazyproperty
    def cube_sum(self):
        """_BaseCubeSum subclass object for this stripe."""
        return _BaseCubeSums.factory(self._cube, self._rows_dimension)

    @lazyproperty
    def unweighted_cube_counts(self):
        """_BaseUnweightedCubeCounts subclass object for this stripe."""
        return _BaseUnweightedCubeCounts.factory(
            self._cube, self._rows_dimension, self._ca_as_0th, self._slice_idx
        )

    @lazyproperty
    def weighted_cube_counts(self):
        """_BaseWeightedCubeCounts subclass object for this stripe."""
        return _BaseWeightedCubeCounts.factory(
            self._cube, self._rows_dimension, self._ca_as_0th, self._slice_idx
        )


class _BaseCubeMeasure(object):
    """Base class for all cube-measure objects."""

    def __init__(self, rows_dimension):
        self._rows_dimension = rows_dimension


# === MEANS ===


class _BaseCubeMeans(_BaseCubeMeasure):
    """Base class for means cube-measure variants."""

    def __init__(self, rows_dimension, means):
        super(_BaseCubeMeans, self).__init__(rows_dimension)
        self._means = means

    @classmethod
    def factory(cls, cube, rows_dimension):
        """Return _BaseCubeMeans subclass instance appropriate to `cube`."""
        if cube.means is None:
            raise ValueError("cube-result does not contain cube-means measure")
        MeansCls = (
            _MrCubeMeans if rows_dimension.dimension_type == DT.MR else _CatCubeMeans
        )
        return MeansCls(rows_dimension, cube.means)

    @lazyproperty
    def means(self):
        """1D np.float64 ndarray of mean for each stripe row."""
        raise NotImplementedError(
            "`%s` must implement `.means`" % type(self).__name__
        )  # pragma: no cover


class _CatCubeMeans(_BaseCubeMeans):
    """Means cube-measure for a non-MR stripe."""

    @lazyproperty
    def means(self):
        """1D np.float64 ndarray of mean for each stripe row."""
        return self._means


class _MrCubeMeans(_BaseCubeMeans):
    """Means cube-measure for an MR stripe.

    Its `.means` is a 2D ndarray with axes (rows, sel/not).
    """

    @lazyproperty
    def means(self):
        """1D np.float64 ndarray of mean for each stripe row."""
        return self._means[:, 0]


# === SUM ===


class _BaseCubeSums(_BaseCubeMeasure):
    """Base class for sum cube-measure variants."""

    def __init__(self, rows_dimension, sums):
        super(_BaseCubeSums, self).__init__(rows_dimension)
        self._sums = sums

    @classmethod
    def factory(cls, cube, rows_dimension):
        """Return _BaseCubeSum subclass instance appropriate to `cube`."""
        if cube.sums is None:
            raise ValueError("cube-result does not contain cube-sum measure")
        SumCls = _MrCubeSums if rows_dimension.dimension_type == DT.MR else _CatCubeSums
        return SumCls(rows_dimension, cube.sums)

    @lazyproperty
    def sums(self):
        """1D np.float64 ndarray of sum for each stripe row."""
        raise NotImplementedError(
            "`%s` must implement `.sum`" % type(self).__name__
        )  # pragma: no cover


class _CatCubeSums(_BaseCubeSums):
    """Sums cube-measure for a non-MR stripe."""

    @lazyproperty
    def sums(self):
        """1D np.float64 ndarray of sum for each stripe row."""
        return self._sums


class _MrCubeSums(_BaseCubeSums):
    """Sums cube-measure for an MR stripe.
    Its `.sums` is a 2D ndarray with axes (rows, sel/not).
    """

    @lazyproperty
    def sums(self):
        """1D np.float64 ndarray of sum for each stripe row."""
        return self._sums[:, 0]


# === STD DEV ===


class _BaseCubeStdDev(_BaseCubeMeasure):
    """Base class for stddev cube-measure variants."""

    def __init__(self, rows_dimension, stddev):
        super(_BaseCubeStdDev, self).__init__(rows_dimension)
        self._stddev = stddev

    @classmethod
    def factory(cls, cube, rows_dimension):
        """Return _BaseCubeStdDev subclass instance appropriate to `cube`."""
        if cube.stddev is None:
            raise ValueError("cube-result does not contain cube-stddev measure")
        StdDevCls = (
            _MrCubeStdDev if rows_dimension.dimension_type == DT.MR else _CatCubeStdDev
        )
        return StdDevCls(rows_dimension, cube.stddev)

    @lazyproperty
    def stddev(self):
        """1D np.float64 ndarray of stddev for each stripe row."""
        raise NotImplementedError(
            "`%s` must implement `.stddev`" % type(self).__name__
        )  # pragma: no cover


class _CatCubeStdDev(_BaseCubeStdDev):
    """StdDev cube-measure for a non-MR stripe."""

    @lazyproperty
    def stddev(self):
        """1D np.float64 ndarray of stddev for each stripe row."""
        return self._stddev


class _MrCubeStdDev(_BaseCubeStdDev):
    """StdDev cube-measure for an MR stripe.
    Its `.stddev` is a 2D ndarray with axes (rows, sel/not).
    """

    @lazyproperty
    def stddev(self):
        """1D np.float64 ndarray of stddev for each stripe row."""
        return self._stddev[:, 0]


# === UNWEIGHTED COUNTS ===


class _BaseUnweightedCubeCounts(_BaseCubeMeasure):
    """Base class for unweighted-count cube-measure variants."""

    def __init__(self, rows_dimension, unweighted_counts):
        super(_BaseUnweightedCubeCounts, self).__init__(rows_dimension)
        self._unweighted_counts = unweighted_counts

    @classmethod
    def factory(cls, cube, rows_dimension, ca_as_0th, slice_idx):
        """Return _BaseUnweightedCubeCounts subclass instance appropriate to `cube`."""
        if ca_as_0th:
            return _CatUnweightedCubeCounts(
                rows_dimension, cube.unweighted_counts[slice_idx]
            )

        if rows_dimension.dimension_type == DT.NUM_ARRAY:
            return _NumArrUnweightedCubeCounts(rows_dimension, cube.unweighted_counts)

        if rows_dimension.dimension_type == DT.MR:
            return _MrUnweightedCubeCounts(rows_dimension, cube.unweighted_counts)

        return _CatUnweightedCubeCounts(rows_dimension, cube.unweighted_counts)

    @lazyproperty
    def bases(self):
        """1D np.float64 ndarray of unweighted table-proportion denonimator per row."""
        raise NotImplementedError(
            "`%s` must implement `.bases`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row."""
        raise NotImplementedError(
            "`%s` must implement `.pruning_base`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def unweighted_counts(self):
        """1D np.float64 ndarray of unweighted-count for each row of stripe."""
        raise NotImplementedError(
            "`%s` must implement `.unweighted_counts`" % type(self).__name__
        )  # pragma: no cover


class _CatUnweightedCubeCounts(_BaseUnweightedCubeCounts):
    """Unweighted-counts cube-measure for a non-MR stripe."""

    @lazyproperty
    def bases(self):
        """1D np.float64 ndarray of table-proportion denonimator (base) for each row.

        Each row in a CAT stripe has the same base (the table-base).
        """
        return np.broadcast_to(self.table_base, self._unweighted_counts.shape)

    @lazyproperty
    def pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row.

        Because this matrix has no MR dimension, this is simply the unweighted count for
        each row.
        """
        return self._unweighted_counts

    @lazyproperty
    def table_base(self):
        """Scalar np.float64 unweighted-N for overall stripe.

        This is the unweighted count of respondents who provided a valid response to
        the question.
        """
        return np.sum(self._unweighted_counts)

    @lazyproperty
    def unweighted_counts(self):
        """1D np.float64 ndarray of unweighted-count for each row of stripe."""
        return self._unweighted_counts


class _MrUnweightedCubeCounts(_BaseUnweightedCubeCounts):
    """Unweighted-counts cube-measure for an MR slice.

    Its `._unweighted_counts` is a 2D ndarray with axes (rows, sel/not).
    """

    @lazyproperty
    def bases(self):
        """1D np.float64 ndarray of table-proportion denonimator (base) for each row.

        Each row in an MR stripe has a distinct base. These values include both the
        selected and unselected counts.
        """
        return np.sum(self._unweighted_counts, axis=1)

    @lazyproperty
    def pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row.

        These values include both the selected and unselected counts of the MR rows
        dimension.
        """
        return np.sum(self._unweighted_counts, axis=1)

    @lazyproperty
    def unweighted_counts(self):
        """1D np.float64 ndarray of unweighted-count for each row of stripe."""
        return self._unweighted_counts[:, 0]


class _NumArrUnweightedCubeCounts(_BaseUnweightedCubeCounts):
    """Unweighted-counts cube-measure for a numeric array stripe."""

    @lazyproperty
    def bases(self):
        """1D np.int64 ndarray of table-proportion denonimator for each cell."""
        return self._unweighted_counts

    @lazyproperty
    def pruning_base(self):
        """1D np.int64 ndarray of unweighted-N for each matrix row."""
        return self._unweighted_counts

    @lazyproperty
    def unweighted_counts(self):
        """1D np.int64 ndarray of unweighted-count for each row of stripe."""
        return self._unweighted_counts


# === WEIGHTED COUNTS ===


class _BaseWeightedCubeCounts(_BaseCubeMeasure):
    """Base class for weighted-count cube-measure variants."""

    def __init__(self, rows_dimension, weighted_counts):
        super(_BaseWeightedCubeCounts, self).__init__(rows_dimension)
        self._weighted_counts = weighted_counts

    @classmethod
    def factory(cls, cube, rows_dimension, ca_as_0th, slice_idx):
        """Return _BaseWeightedCubeCounts subclass instance appropriate to `cube`."""
        if ca_as_0th:
            return _CatWeightedCubeCounts(rows_dimension, cube.counts[slice_idx])

        if rows_dimension.dimension_type == DT.MR:
            return _MrWeightedCubeCounts(rows_dimension, cube.counts)

        return _CatWeightedCubeCounts(rows_dimension, cube.counts)

    @lazyproperty
    def bases(self):
        """1D np.float64 ndarray of table-proportion denominator for each cell."""
        raise NotImplementedError(
            "`%s` must implement `.bases`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def table_margin(self):
        """Scalar or 1D np.float64 array of weighted-N for overall stripe."""
        raise NotImplementedError(
            "`%s` must implement `.table_margin`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def weighted_counts(self):
        """1D np.float64 ndarray of weighted-count for each row of stripe.

        When the cube-result has no weight these values are the same as the
        unweighted-counts.
        """
        raise NotImplementedError(
            "`%s` must implement `.weighted_counts`" % type(self).__name__
        )  # pragma: no cover


class _CatWeightedCubeCounts(_BaseWeightedCubeCounts):
    """Weighted-counts cube-measure for a non-MR stripe.

    Its `._weighted_counts` is a 1D ndarray with axes (rows,).
    """

    @lazyproperty
    def bases(self):
        """1D np.float64 ndarray of table-proportion denominator for each cell."""
        return np.broadcast_to(self.table_margin, self._weighted_counts.shape)

    @lazyproperty
    def table_margin(self):
        """Scalar np.float64 weighted-N for overall stripe.

        This is the weighted count of respondents who provided a valid response to
        the question.
        """
        return np.sum(self._weighted_counts)

    @lazyproperty
    def weighted_counts(self):
        """1D np.float64 ndarray of weighted-count for each row of stripe.

        When the cube-result has no weight these values are the same as the
        unweighted-counts.
        """
        return self._weighted_counts


class _MrWeightedCubeCounts(_BaseWeightedCubeCounts):
    """Weighted-counts cube-measure for an MR slice.

    Its `._weighted_counts` is a 2D ndarray with axes (rows, sel/not).
    """

    @lazyproperty
    def bases(self):
        """1D np.float64 ndarray of table-proportion denominator for each cell."""
        # --- (weighted) bases for an MR slice is the already 1D table-margin ---
        return self.table_margin

    @lazyproperty
    def table_margin(self):
        """1D np.float64 weighted-N for each row of stripe.

        This is the weighted count of respondents who provided a valid response to the
        question. Both selecting and not-selecting the subvar/option are valid
        responses, so this value includes both the selected and unselected counts.
        """
        return np.sum(self._weighted_counts, axis=1)

    @lazyproperty
    def weighted_counts(self):
        """1D np.float64 ndarray of weighted-count for each row of stripe."""
        return self._weighted_counts[:, 0]
